; QB Copyright (C)2001 Andrew Davie
; see http://homepages.ihug.co.nz/~aaronat/qb.html

; DO NOT TRUST ABSOLUTELY THE COMMENTS IN THIS CODE!
; MOST OF THEM ARE OK.  SOME ARE MISLEADING :)

; History: original PUSH demo posted to Stella list May 1998
; This demonstrated multiple-colour playfields and VERY rudimentary background animation

; Extended January/February 2001 to display a 4x4 tiled grid, each tile independantly movable.


; Known bugs... (but mostly, who cares)
;       Rabbit doesn't face always correct direction when shifting cubes (?)
;       Fireball sometimes just halts (doesn't hunt)
;       Small timing/screen glitch when fireball jumps to bottom row (sometimes)


        processor 6502
        include vcs.h

;------------------------------------------------------------------------------
; Compilation switches...

COMPILE_ILLEGALOPCODES  = 1             ;             set to 0 for NO illegal opcodes, 1 for YES illegal opcodes
        ;\--> Users of Stella (MAC users) need to set the above to 0 (for now, anyway)

NTSC            = 0
PAL             = 1

COMPILE_VERSION         = PAL           ; NTSC/PAL    switches display and colours
TITLESCREEN             = 0             ; include code for title screen
TESTMAXCUBE             = 0             ; test maximum cubes (for timing)

FORCEHUNTER             = 0             ; force the fireball (for testing)
FORCESHIFTER            = 0             ; force the octopus (for testing)
FORCEHELPER             = 0             ; force the rabbit (for testing)

; Possible savings - set the following to 0 to disable (and save memory!)

PLAYER_HAS_HAIR         = 0             ; [ 12 bytes] gives player interlaced hair [22 if rising]
PLAYER_WEARS_HAT        = 0             ; [     free] dodgy colour change for hat                      NOT IMPLEMENTED
ATTRACT_MODE            = 1             ; [ 13 bytes] cycle colours during demo
DIAGONALS               = 0             ; diagonals (unimplemented)

; Gameplay tweaking...

TIME_PAUSE_AFTER_JUMP   = 0 ;1             ; max 31

PLAYERCOLOUR            = COLUP0        ; COLUP0 or COLUBK (lets you see colour timing)
PLAYER0SHAPE            = GRP0          ; PLAYER0SHAPE or COLUBK "
PLAYER1SHAPE            = GRP1          ; PLAYER1SHAPE or COLUBK "

;------------------------------------------------------------------------------
; THINGS YOU DON'T CHANGE...

 
CUBEXY          = 5                     ; size of cube + border
PFSIZE          = 3*(4*CUBEXY-1)        ; 3 bytes / line by n lines
PFC             = 19                    ; actual rows of playfield

NPLAYERS        = 2                     ; DO NOT CHANGE

SPRITE_Y_ADJUST = 24                    ; adjust vertical position of sprites (pixels down from top of cube)
SPRITE_X_ADJUST = 21                    ; horizontal pixel adjust to align with grid

GUARANTEED0READ = MTA_ZERO              ; ANY zero page locations with 0,0 in consecutive bytes


NUMBEROFCUBES   = 6                     ; cubes which move
 ;\---> see dependencies :
 ;       - PlayerCube assumes <= 8
 ;       - uses 2n bytes RAM
 ;      NOT MUCH POINT >8 CUBES, AS IT GETS CLUTTERED/EASY!


SPRITE_HEIGHT   = 11
 


;------------------------------------------------------------------------------
; Generic equates

; bits
D0              = 1
D1              = 2
D2              = 4
D3              = 8
D4              = 16
D5              = 32
D6              = 64
D7              = 128

BLANK_LINE      = 0                     ; header for all sprite frames

;------------------------------------------------------------------------------
; variables

    SEG.U vars
    org $80

    ; only 128 bytes... including the stack!
    ; let the assembler calculate the addresses

PlayerCounter   ds NPLAYERS             ; generic counter for event timing
                ; mmmccccc
                ; m = creature mode
                ; c = creature counter
        ;\--> mode indicates what to do
        ;     counter is used for THAT mode only

        ;\--> usage/mode definitions
MASK_MODE       = %11100000
MODE_INIT       = %00000000                     ; init state for creatures
MODE_NORMAL     = %00100000                     ; doing nothing special
MODE_TOSSOFF    = %01000000                     ; tossed off screen
MODE_JUMP       = %01100000                     ; jumping
;MODE_LAND       = %10000000                     ; landed (possible pause)
MODE_HATCH      = %01100000

;MODE_LURK       = %10000000                     ; lurking, waiting for a cube
;MODE_JUMP       = %10100000                     ; jumping
;MODE_LOCK       = %11000000                     ; locked, pending screen change
;MODE_DELAY      = %11100000                     ; just delays

; These must correspond to the correct entries in TypeVector offset tables for each creature

MODE_PLAYER_LOCK        = %10100000
MODE_PLAYER_KILL        = %11000000
MODE_DEATHWATCH         = %11100000


CubeX           ds NUMBEROFCUBES        ; d0-d4 = X(0-31), high 3 bits unused.

                ; v--xxxxx
                ; v = valid cube (1 = invalid)
                ; x = x coordinate (0-31).  16+ is off board
                ; - = unused


CubeYDir        ds NUMBEROFCUBES
                ; sdddyyyy
                ; y = Y coordinate (0-15)
                ; d = direction:
                ;       000     up
                ;       001     right
                ;       010     down
                ;       011     left
                ;       100     upright
                ;       101     downright
                ;       110     downleft
                ;       111     upleft
                ; s = stationary (1 = stopped, 0 = moving)


rnd             ds 2                    ; random #
timer           ds 2                    ; generic 1-per-frame timer + screen rate (byte 2)

PlayerFlags     ds NPLAYERS             ; generic flags
                ; dddmfffx
                ; d = direction of moving/facing. x00=up, x01=right, x10=down, x11=left
                        ;\-->being extended for diagonals... in progress
                ; f = frame to display for animation
                ; x = show lives instead of score
                ; m = mirror sprite (left-facing)
        ;\--> usage definitions
FLAG_MIRROR     = D4

MASK_FRAME              = %00001110
MASK_DIRECTION          = %11100000

FRAMEPLUSONE            = %00000010

FRAME_JUMP              = %00000010                     ; jump frame
FRAME_STAND             = %00000000                     ; standing frame
FRAME_NOPE              = %00000100                     ; no go (poof!)
FRAME_SAD               = %00000100
FRAME_BLINK             = %00000110

DISPLAY_LIVES_SCORE     = %00000001



PlayerX         ds NPLAYERS             ; x coordinate
PlayerY         ds NPLAYERS             ; y offset for jumping, etc.

PlayerCube      ds NPLAYERS
                ; values 0-NUMBEROFCUBES-1
                ; Let's assume NUMBEROFCUBES <= 8  (OK, so that's only 3 bits required)
                ; bddddiii
                ; i, d0-d2 = cube index
                ; d, d3-d6 = sprite/creature ID (used for graphic selection) (16 objects)
                ; b, d7    = bonus not available, decrement score !!
        ;\-->usage...
MASK_TYPE       = %01111000
MASK_CUBE       = %00000111



TARGETSIZE = 5
TargetDisplay   ds 2*TARGETSIZE         ; visible target display (flashes/blinks, etc)
                ; TargetDisplay hold the graphics for the entire right-side of the screen
                ; The first 4 bytes in each are the target area, which is flashed/animated
                ; The last byte is (all of!) the sliding cube feeding onto the grid.

Grid            ds 4                    ; 4 rows.  2 bits/tile for info
                ; GRID: organised into 4 rows, 1 byte/row
                ;       each byte has 2 bits/tile, ordered target/ownership
                ;       target = a cube 'belongs' here
                ;       ownership = a cube IS here


playfield       ds 3 * 19
        ; displays on the screen horizontally, but organised in COLUMNS, each PFC bytes deep
        ; there are 3 columns.  Organisation this way saved nearly 50 bytes of code!

level           ds 1                    ; game level (increases difficulty)
        ; mmmlllll
        ; mmm = # of lives (0-7)
        ; lllll = level (0-31)
        ;\-> usage...
MASKLEVEL       = %11111
MASKLIVES       = %11100000
LIVESPLUSONE    = %00100000
LIVESMAX        = %11100000             ; maximum lives
ONE_LIFE        = %00100000             ; one life +


Cubes           ds 1                    ; number of cubes on this board
        ; sdfffccc
        ; s = grid complete, flash!
        ; d = demo mode
        ; f = fruit progression counter
        ; c = remaining cube count
        ;\-->equates...
CUBES_DEMO_MODE equ %01000000           ; game is in demo mode
CUBES_COMPLETED equ %10000000           ; grid has completed            (optimised to assume D7 in code)

MASK_FRUIT              = %00011000     ; only uses 2 bits (0-3 values)
FRUITPLUSONE            = %00001000


BackColour      ds 1                    ; background colour (mainly debugging)

PlayerScore     ds 3                    ; player's score (3 BCD digits)

        ; /DODGY ON
        ; This overlay variable is used for the overlay variables.  That's OK.
        ; However, it is positioned at the END of the variables so, if on the off chance we're overlapping
        ; stack space and variable, it is LIKELY that that won't be a problem, as the temp variables
        ; (especially the latter ones) are only used in rare occasions.

        ; FOR SAFETY, DO NOT USE THIS AREA DIRECTLY (ie: NEVER reference 'Overlay' in the code)
        ; ADD AN OVERLAY FOR EACH ROUTINE'S USE, SO CLASHES CAN BE EASILY CHECKED


Overlay         ds 19                   ;--> OVERLAYS

        ; /DODGY OFF

;------------------------------------------------------------------------------
; OVERLAYS!
; These variables are overlays, and should be managed with care
; That is, variables are ALREADY DEFINED, and we're reusing RAM for other purposes

; EACH OF THESE ARE VARIABLES (TEMPORARY) USED BY ONE ROUTINE (AND IT'S SUBROUTINES)
; THAT IS, LOCAL VARIABLES.  USE 'EM FREELY, THEY COST NOTHING

; TOTAL SPACE USED BY ANY OVERLAY GROUP SHOULD BE <= SIZE OF 'Overlay'

;------------------------------------------------------------------------------
; MTarget(Area) overlays...
        
        org Overlay ;[19/19]

SpriteEnd       ds 2
Sprite          ds 4                    ; -> sprite graphic shape
ColourPlayer    ds 2                    ; player 0 colour

; DrawScreen overlays...

DS_Sprite       ds 2
DS_temp         ds 1
MT_Overlay      ds 8                    ;--> OVERLAYS

; SUBSETS of DrawScreen
; PositionSprites overlays...

        org MT_Overlay ;[1/8]

PS_temp         ds 1

; MTargetAreaBorder overlays...

        org MT_Overlay ;[8/8]

; MTargetArea overlays

MTA_TargBorder  ds 1

MTA_ZERO        ds 2                    ; BRILLIANT!  gives us our two required 0-reads
MTA_31          ds 2

MT_TargetMask   ds 1                    ; mask for target area border (SHARED)

MTA_TargetL     ds 1                    ; calculated target-area display bytes
MTA_TargetR     ds 1


;------------------------------------------------------------------------------
; ScoreLine overlays...
; OVERLAP WITH DRAWSCREEN... BE *VERY* CAREFUL NOT TO BRACKET USAGE
; ACROSS SCORELINE!

        org Overlay ;[54/19]

S0              ds 2                    ; used for addressing digits of score
S1              ds 2
S2              ds 2
S3              ds 2
S4              ds 2
S5              ds 2

stkp            ds 1
sreg            ds 1
loop            ds 1


;------------------------------------------------------------------------------
; DoAI overlays... 

        org Overlay ;[12/19]

AI_Player       ds 1
AI_temp         ds 1
AI_cube         ds 1
AI_cube2        ds 1
AI_Overlay      ds 4                    ;--> OVERLAYS

; SUBSETS of DOAI
; CountDown overlays...

CD_TimerMask    ds 1

; AI_Jump overlays...
AIJ_YSlower     ds 1

; AI_Lurk overlays...
AIL_LurkerCube  ds 1

; NextLevel overlays...
NL_temp         ds 1

; SUB-OVERLAYS OF DoAI...(!)
; OVERLAYS INSIDE OVERLAYS

        org AI_Overlay ;[4/4]
; CheckMove (sub)overlays...
CM_CubeIndex    ds 1
CM_Direction    ds 1
CM_Column       ds 1
CM_scratch      ds 1

        org AI_Overlay ;[2/4]
AI_Vector       ds 2

        org AI_Overlay ;[4/4]
; CheckJump (sub)overlays...
CJ_CubeIndex    ds 1
CJ_Direction    ds 1
CJ_Overlay      ds 2                    ; --> OVERLAYS

        org CJ_Overlay ;[2/2]
CJ_NewX         ds 1
CJ_NewY         ds 1

        org CJ_Overlay ;[1/2]
CJ_CubeNumber   ds 1

;------------------------------------------------------------------------------
; InitCubes overlays...

        org Overlay ;[8/19]

IC_FedCubeIndex ds 1

; SUBSETS OF InitCubes...
; CreateTableaux overlays...
CT_Cubes        ds 1
CT_index        ds 1
CT_coord        ds 1

; FeedCube overlays...
FC_FedCubeIndex ds 1
FC_BlankSpot    ds 1
FC_CubeX        ds 1
FC_CubeY        ds 1
FC_temp2        ds 1
FC_coord        ds 1

;------------------------------------------------------------------------------
; MoveCube overlay...

        org Overlay ;[8/19]

MC_temp         ds 1
MC_DirOnly      ds 1
MC_HasMoved     ds 1
MC_Complete     ds 1

; SUBSETS OF MoveCube...
; DrawCube overlay...

DC_scratch      ds 1
DC_CubeLine     ds 1
DC_saveX        ds 1


;------------------------------------------------------------------------------

        ; add more overlays here, as required...



;------------------------------------------------------------------------------
; Alternates...  these variables used in the title screen
; The title screen calls ScoreLine, so requires the Overlay area to be available for that

        org $80

TS_line         ds 1
TS_width        ds 1
TS_temp         ds 1

;------------------------------------------------------------------------------


        SEG code
        org $F000
; eek! just 4k (-vectors)


;------------------------------------------------------------------------------

PSI     .byte 8,4 ;,0  << merge 0

        ; merged...

PreDrawn3

        REPEAT 17
        .byte 0
        REPEND

    .byte %10000000
    .byte %11000000
    .byte %11100000

;\--> merges with next table... don't move 'em!        

PreDrawn0

    ;     pf1 -->   pf2 <--   pf0 <--   pf1 -->   pf2 <--
    ;     ok        flip      flip xxxx ok        flip
    .byte %11110000     ; 0
    .byte %01111000; 1
    .byte %00111100; 2
    .byte %00011110; 3
    .byte %00001111; 4
    .byte %00000111; 5
    .byte %00000011; 6
    .byte %00000001; 7

PreDrawn4

;\--> merges with next table... don't move 'em!        

SSNull                                  ; << below also used for blank frame :)
        REPEAT 16
        .byte 0
        REPEND

;\--> merges with next table... don't move 'em!        

PreDrawn2

        REPEAT 4
        .byte 0
        REPEND
        
;\--> merges with next table... don't move 'em!        

PreDrawn1

    ;     pf1 -->   pf2 <--   pf0 <--   pf1 -->   pf2 <--
    ;     ok        flip      flip xxxx ok        flip
    .byte %00000000
    .byte %00000000
    .byte %00000000
    .byte %00000000
    .byte %00000000
    .byte %00000001
    .byte %00000011
    .byte %00000111
    .byte %00001111
    .byte %00011110
    .byte %00111100
    .byte %01111000
    .byte %11110000
    .byte %11100000
    .byte %11000000
    .byte %10000000

        REPEAT 16 - 8
        .byte 0
        REPEND

;\--> merges with next table... don't move 'em!        

; digits shouldn't cross page boundary...

DigitNull

        .byte 0
        .byte 0
        .byte 0
        .byte 0
        .byte 0
        .byte 0
        .byte 0
        .byte 0

Digit0  .byte %00111100         ;    ****  
        .byte %01001110         ;   *  ***
        .byte %10000110         ;  *    **
        .byte %10000110         ;  *    **
        .byte %10000110         ;  *    **
        .byte %10000110         ;   *  **
        .byte %01001100         ;    ***
        .byte %00111000         ;  

Digit1  .byte %00001100         ;      **
        .byte %00001100         ;      **
        .byte %00001100         ;      **
        .byte %00001100         ;      **
        .byte %00001100         ;      **
        .byte %00101100         ;    * **
        .byte %00011100         ;     ***
        .byte %00001100         ;      **

Digit2  .byte %11111110         ;  *******
        .byte %11000010         ;  **    *
        .byte %01110000         ;   ***
        .byte %00011100         ;     ***
        .byte %00000110         ;       **
        .byte %00000110         ;       **
        .byte %10001100         ;  *   **
        .byte %01111000         ;   ****

Digit3  .byte %01111100         ;   *****
        .byte %10000110         ;  *    **
        .byte %00000110         ;       **
        .byte %00000110         ;       **
        .byte %00011100         ;     ***
        .byte %00001000         ;      *
        .byte %00000100         ;       *
        .byte %01111110         ;   ******

Digit4  .byte %00001100         ;      **
        .byte %00001100         ;      **
        .byte %11111110         ;  *******
        .byte %10001100         ;  *   **
        .byte %01001100         ;   *  **
        .byte %00101100         ;    * **
        .byte %00011100         ;     ***
        .byte %00001100         ;      **

Digit5  .byte %01111100         ;   *****
        .byte %10000110         ;  *    **
        .byte %00000110         ;       **
        .byte %00000110         ;       **
        .byte %00001100         ;      **
        .byte %01111000         ;   ****
        .byte %01100000         ;   **
        .byte %01111110         ;   ******

Digit6  .byte %01111100         ;   *****
        .byte %11000110         ;  **   **
        .byte %11000110         ;  **   **
        .byte %11000110         ;  **   **
        .byte %11111100         ;  ******
        .byte %01100000         ;   **
        .byte %00110000         ;    **
        .byte %00001100         ;      **

Digit7  .byte %00110000         ;    **
        .byte %00110000         ;    **
        .byte %00110000         ;    **
        .byte %00110000         ;     **
        .byte %00011000         ;      **
        .byte %00001100         ;  *    **
        .byte %10000110         ;   ******
        .byte %01111110         ;  

Digit8  .byte %01111100         ;   *****
        .byte %11000110         ;  **   **
        .byte %11000110         ;  **   **
        .byte %11000110         ;  **   **
        .byte %01111000         ;   ****
        .byte %01001100         ;   *  **
        .byte %01001100         ;   *  **
        .byte %00111000         ;    ***

Digit9  .byte %00111000         ;    ***
        .byte %00001100         ;      **
        .byte %00000110         ;       **
        .byte %01111110         ;   ******
        .byte %11000110         ;  **   **
        .byte %11000110         ;  **   **
        .byte %01000100         ;   *   *
        .byte %00111000         ;    ***

        .byte %11101110         ;  *** ***              board indicator
        .byte %11101110         ;  *** ***
        .byte %11101110         ;  *** ***
        .byte %00000000         ;  
        .byte %11101110         ;  *** ***
        .byte %11101110         ;  *** ***
        .byte %11101110         ;  *** ***
        .byte %00000000         ;  

        .byte %01110110         ;   *** **              player indicator
        .byte %00111100         ;    ****  
        .byte %11100010         ;  ***   * 
        .byte %11011000         ;  ** **
        .byte %11111110         ;  *******
        .byte %01101010         ;   ** * *
        .byte %01111100         ;   *****
        .byte %00111000         ;    ***


;------------------------------------------------------------------------------

OwnershipBit    .byte %00000001,%00000100,%00010000,%01000000
TargetBit       .byte %00000010,%00001000,%00100000,%10000000
TargetOwnerBit  .byte %00000011,%00001100,%00110000,%11000000

;------------------------------------------------------------------------------

TargetIndex

        ; Points to the correct line in the TargetDisplay playfield
        ; Indexed by the scanline (0-18) being drawn.

        ; This allows the single kernel to draw the entire playfield (!)

        .byte TargetDisplay+8      ; 18            "
        .byte TargetDisplay+8      ; 17            "
        .byte TargetDisplay+8      ; 16            "
        .byte TargetDisplay+8      ; 15            cube sliding on bottom
        .byte MT_TargetMask        ; 14            "
        .byte MT_TargetMask        ; 13            "
        .byte MT_TargetMask        ; 12            "
        .byte MT_TargetMask        ; 11            "
        .byte MT_TargetMask        ; 10            "
        .byte MTA_31               ; 9             blank
        .byte TargetDisplay+6      ; 8             bottom row target
        .byte TargetDisplay+6      ; 7             bottom row target
        .byte TargetDisplay+4      ; 6             bottom middle row target
        .byte TargetDisplay+4      ; 5             bottom middle row target
        .byte TargetDisplay+2      ; 4             top middle row target
        .byte TargetDisplay+2      ; 3             top middle row target
        .byte TargetDisplay+0      ; 2             top row target
        .byte TargetDisplay+0      ; 1             top row target
        .byte MTA_31               ; 0             blank

MTargetMaskInit
        .byte 255,255,255,255,0,0,0,0,0,0,15,15,15,15,15,15,15,15,0

;------------------------------------------------------------------------------

        IF COMPILE_VERSION = PAL

GREEN           = %00110000
YELLOW          = %00100000
RED             = %01100000
ORANGE          = %00100000

        ELSE

GREEN           = %11100000
YELLOW          = %00100000
RED             = %01000000
ORANGE          = %00100000
        ENDIF

DARK            = %0000
LIGHT           = %1111
MEDIUM          = %1000


;               PAL             NTSC
; 0000
; 0001          grey            light yellow
; 0010          orange          orange
; 0011          lime green      peach
; 0100          orange          pink
; 0101          blue green      purple
; 0110          pink red        dark blue
; 0111          aqua            blue
; 1000          purple          blue
; 1001          aqua            blue
; 1010          purple          aqua
; 1011          blue            green/blue
; 1100          blue purple     melon green
; 1101          dk.blue         olive
; 1110          grey            lt.green
; 1111          grey            orangy


;------------------------------------------------------------------------------

SpriteVector

        ; indexed by creature ID/type
        ; points to the sprite shape/definition(s) for each creature
        ; THESE MUST CORRESPOND TO CREATURE TYPES!

        .word HumanSprite
        .word SPLurker                          ; lurker
        .word SPFruit
        .word SPPoints
        .word SPShifter                         ; hinders
        .word SPTwoShoes                        ; assists
        .word SPHunter                          ; hunts
        .word HumanSprite                       ; corpse
        .word SPPOOF                            ; creature disappearing

;------------------------------------------------------------------------------


; Sprite Graphics...

        ; SPRITE DATA MUST *NOT* CROSS PAGE BOUNDARY!!

; Format:

        ; OLD DOC... WRONG
        ; byte          meaning
        ; 0             ALWAYS 0        used for blank display
        ; 1             vector (offset) to 2nd shape (can be 0, indicating THIS shape)
        ;               \--> allows multi-colour sprites by using 2 shapes, 2 colours
        ; add the vector if needed, and the offsets from THAT point are...
        ; 0             COLOUR of this frame
        ; ...           frame definitions
        ; ONLY for the 1st shape, 10 '0' bytes, used for rising out of cubes
        ; if it does, then they ARE needed for each multiplex colour/shape definition


SPLurker
        .word SSNull
        .word SSLurkEgg1
        .word SSLurkEgg2
        .word SSLurkEgg3
        .word SSLurkEgg3
        .word SSLurkEgg4
 


        IF TITLESCREEN

                ; merged...

QbTitle

        ; PF0=0, PF1<-, PF2->, PF0 (4 bits), PF1, PF2[4,4 blank]
        ; 2nd PF0 embedded in 2nd PF2

;        .byte 0                relies on merge with above 0-byte
        .byte %00000011
        .byte %00001111
;        .byte %00011111
;        .byte %00111110
;        .byte %01111100
;        .byte %01111000
;        .byte %01111000
;        .byte %01111000
;        .byte %01111000
;        .byte %01111000
;        .byte %01111100
;        .byte %00111110
;        .byte %00011111
;        .byte %00001111
;        .byte %00000011

        ENDIF

SSLurkEgg2

        .byte 0
        .byte %00101111

        .byte %00000000  
        .byte %00000000  
        .byte %00000000  
        .byte %00111100  
        .byte %01111110  
        .byte %01111110  
        .byte %01111110  
        .byte %01111110  
        .byte %01111110  
        .byte %00111100  
        .byte BLANK_LINE


        IF TITLESCREEN
                ; merged...

QbTitle1

        ; PF0=0, PF1<-, PF2->, PF0 (4 bits), PF1, PF2[4,4 blank]
        ; 2nd PF0 embedded in 2nd PF2


;        .byte 0                relies on 0-byte, above
;        .byte %10001111
;        .byte %11011111
;        .byte %11001111
;        .byte %11111010
;        .byte %11111110
;        .byte %01111100
;        .byte %11000000
;        .byte %11000000
;        .byte %11000000
;        .byte %11000000
;        .byte %11100000
;        .byte %11110000
;        .byte %11111111
;        .byte %01111111
;        .byte %00011111
        ENDIF


SSLurkEgg4

        .byte 0
        .byte %00101111

        .byte %00111100  
        .byte %01111110  
        .byte %01111110  
        .byte %11111011  
        .byte %01100110  
        .byte %10001001  
        .byte %11111111  
        .byte %11111111  
        .byte %01111110  
        .byte %00111100  
SSLurkEgg3

        .byte BLANK_LINE

        ; merged...

;        .byte 0
        .byte %00101111

        .byte %00111100  
        .byte %01111110  
        .byte %01111110  
        .byte %11111111  
        .byte %01111110  
        .byte %10011111  
        .byte %11111111  
        .byte %11111111  
        .byte %01111110  
        .byte %00111100  
SSPOOF

        .byte BLANK_LINE

        ; merged...
;        .byte 0
        .byte %00101111

        .byte %00000000 ;
        .byte %00000000 ;
        .byte %01000010 ; *    *
        .byte %00100100 ;  *  *
        .byte %00011000 ;   **
        .byte %11100111 ;***  ***
        .byte %00011000 ;   **
        .byte %00100100 ;  *  *
        .byte %01000010 ; *    * 
        .byte %00000000 ;

                ; merged...

DirShifted

        .byte BLANK_LINE


        ;.byte 0
        .byte %00010000,%00100000,%00110000



        IF TITLESCREEN
                ; merged...

QbTitle2

        ; PF0=0, PF1<-, PF2->, PF0 (4 bits), PF1, PF2[4,4 blank]
        ; 2nd PF0 embedded in 2nd PF2

;        .byte 0                relies on 0-byte, above
        .byte %11110001
        .byte %01110011
        ;.byte %10110111
        ;.byte %11011110
        ;.byte %10101110
        ;.byte %10111110
        ;.byte %10111110
        ;.byte %10111111
;        .byte %10110111
;        .byte %10110011
;        .byte %10110000
;        .byte %10010000
;        .byte %10000000
;        .byte %11000000
;        .byte %11100000
        ENDIF



SS_Blink
      
        .byte 0
        
        .byte %00001100                 ; white

        .byte %00111100                 ; actual data
        .byte %01111110
        .byte %01111111
        .byte %11111111
        .byte %11011110
        .byte %11100001
        .byte %00111110
        .byte %01101100
        .byte %01000100
        .byte %01110110

                ; merged...


SS_Normal
      
        .byte BLANK_LINE
        
        .byte %00001100                 ; white

        .byte %00111100                 ; actual data
        .byte %01111110
        .byte %01110101
        .byte %11111111
        .byte %11011110
        .byte %11100001
        .byte %00111110
        .byte %01101100
        .byte %01000100
        .byte %01110110

                ; merged...

SS_Jump
        .byte BLANK_LINE

;        .byte 0                 ; no multiplex offset
        .byte %00001100                 ; white

        .byte %00000000
        .byte %00111100                 ; actual data
        .byte %01111110
        .byte %01110101
        .byte %11111111
        .byte %11011000
        .byte %11100001
        .byte %00111110
        .byte %01110011

                ; merged...

BlockIndex

        .byte %00000000
        .byte BLANK_LINE


        ; Determines WHAT is to be displayed on the lines

        .byte %00000000
        .byte %10000000,%11111111,%11110000



        IF TITLESCREEN
                ; merged...  THIS NO LONGER VALID

QbTitle3

        ; PF0=0, PF1<-, PF2->, PF0 (4 bits), PF1, PF2[4,4 blank]
        ; 2nd PF0 embedded in 2nd PF2

;        .byte 0                        relies on 0-byte, above
        .byte %10111111
        .byte %11011111
        .byte %00011001
        .byte %11100000
        .byte %11100000
        .byte %11100000
        .byte %11100000
        .byte %11111000
        .byte %11111111
        .byte %11110111
        .byte %11100000
        .byte %11100000
        .byte %11100000
        .byte %11000000
        .byte %10000000
        ENDIF


SS_Sad
      
        .byte 0
        
        .byte %00001100                 ; white

        .byte %00111100                 ; actual data
        .byte %01111110
        .byte %01110101
        .byte %11111111
        .byte %11110000
        .byte %11101111
        .byte %00111110
        .byte %01101100
        .byte %01000100
        .byte %01110110
        .byte BLANK_LINE

;------------------------------------------------------------------------------

SSFruit_Apple

        ; This is a multicolour sprite
        ; OK 17/2

        .byte 12                ; offset to 2nd shape

        IF COMPILE_VERSION = PAL
                .byte %01100111                ; green rind
        ELSE
                .byte %01001000
        ENDIF

;               PAL             NTSC
; 0000
; 0001          grey            light yellow
; 0010          orange          orange
; 0011          lime green      peach
; 0100          orange          pink
; 0101          blue green      purple
; 0110          pink red        dark blue
; 0111          aqua            blue
; 1000          purple          blue
; 1001          aqua            blue
; 1010          purple          aqua
; 1011          blue            green/blue
; 1100          blue purple     melon green
; 1101          dk.blue         olive
; 1110          grey            lt.green
; 1111          grey            orangy

        .byte %00000000
        .byte %00000000
        .byte %00000000
        .byte %01111110
        .byte %11111111
        .byte %11111111
        .byte %11111111
        .byte %01111110
        .byte %00111100
        .byte %00000000
        .byte BLANK_LINE

        IF COMPILE_VERSION = PAL
                .byte %00110111                ; green rind
        ELSE
                .byte %11000111
        ENDIF

        .byte %11100110
        .byte %01101110
        .byte %00010000
        .byte %01111110
        .byte %11110111
        .byte %11111111
        .byte %11111111
        .byte %01111110
        .byte %00111100
        .byte %00000000
        .byte BLANK_LINE


SSFruit_Heart

        ; A chalice... this game's HOLY GRAIL, I guesss :)

        ; This is a multicolour sprite
;        ; OK 17/2

        .byte 0 ;12                ; offset to 2nd shape

        .byte %00101110                 ; gold

        .byte %11111111
        .byte %11111101
        .byte %11111101
        .byte %01111010
        .byte %01111010
        .byte %00111100
        .byte %00011000
        .byte %00111100
        .byte %00011000
        .byte %01111110
        .byte BLANK_LINE

;        .byte %00101110                ; green rind

;        .byte %11110001
;        .byte %11111101
;        .byte %11111101
;        .byte %01111010
;        .byte %01111010
;        .byte %00111100
;        .byte %00011000
;        .byte %00110000
;        .byte %00011000
;        .byte %01111110
;        .byte BLANK_LINE





SSFruit_Watermelon

        ; This is a multicolour sprite
        ; OK 17/2

        .byte 12                ; offset to 2nd shape

        IF COMPILE_VERSION = PAL
                .byte %01100110                 ; red flesh
        ELSE
                .byte %01000111
        ENDIF

        .byte %00000000
        .byte %00000110
        .byte %00011110
        .byte %00111110
        .byte %00111110
        .byte %00111110
        .byte %00111110
        .byte %00011110
        .byte %00000110
        .byte %00000000
        .byte BLANK_LINE

        IF COMPILE_VERSION = PAL
                .byte %00110111                ; green rind
        ELSE
                .byte %11001000
        ENDIF

        .byte %00000110
        .byte %00011000
        .byte %00100100
        .byte %01001000
        .byte %01100100
        .byte %01000000
        .byte %01010000
        .byte %00100100
        .byte %00011000
        .byte %00000110
        .byte BLANK_LINE

SSFruit_Pineapple 

        .byte 12

        IF COMPILE_VERSION = PAL
                .byte %01100111                 ; red flesh
        ELSE
                .byte %01001010                 ; red flesh
        ENDIF
;               PAL             NTSC
; 0000
; 0001          grey            light yellow
; 0010          orange          orange
; 0011          lime green      peach
; 0100          orange          pink
; 0101          blue green      purple
; 0110          pink red        dark blue
; 0111          aqua            blue
; 1000          purple          blue
; 1001          aqua            blue
; 1010          purple          aqua
; 1011          blue            green/blue
; 1100          blue purple     melon green
; 1101          dk.blue         olive
; 1110          grey            lt.green
; 1111          grey            orangy


        .byte %00000000
        .byte %00000000
        .byte %00000000
        .byte %00111100
        .byte %01111110
        .byte %01111110
        .byte %01111110
        .byte %01111110
        .byte %01111110
        .byte %00111100
        .byte BLANK_LINE

        IF COMPILE_VERSION = PAL
                .byte %00110111                ; green leaves
        ELSE
                .byte %11101010
        ENDIF

        .byte %01010111
        .byte %10111001
        .byte %00011010
        .byte %00111100
        .byte %01011110
        .byte %01110110
        .byte %01011010
        .byte %01101110
        .byte %01010110
        .byte %00111100
        .byte BLANK_LINE



SSShifter

        ; Octopus graphic

        .byte 12


        IF COMPILE_VERSION = PAL
                .byte %11001000
        ELSE
                .byte %10101000
        ENDIF

        .byte %00111100 ;    ****
        .byte %01111110 ;   ******
        .byte %11001001 ;  **  *  *
        .byte %11101011 ;  *** * **
        .byte %11111111 ;  ********
        .byte %01111110 ;   ******
        .byte %00111110 ;    *****
        .byte %11111111 ;  ********
        .byte %00111110 ;    *****
        .byte %01001001 ;   *  *  *

        .byte BLANK_LINE

        IF COMPILE_VERSION = PAL
                .byte %11001001                ; greenish
        ELSE
                .byte %10101001
        ENDIF

        .byte %00111100 ;    ****
        .byte %01111110 ;   ******
        .byte %11111111 ;  ********
        .byte %11001001 ;  **  *  *
        .byte %11101011 ;  *** * **
        .byte %01111110 ;   ******
        .byte %00011100 ;     ***
        .byte %11000011 ;  **    **
        .byte %00101010 ;    * * *
        .byte %01001001 ;   *  *  *

        .byte BLANK_LINE

SPFruit
        .word SSFruit_Watermelon
        .word SSFruit_Pineapple
        .word SSFruit_Apple
        .word SSFruit_Heart

SPPoints
        .word SSPoints_100
        .word SSPoints_300
        .word SSPoints_500
        .word SSPoints_1UP

SPShifter
        .word SSShifter         ; stand
        .word SSShifter         ; jump

; 0000
; 0001          grey            light yellow
; 0010          orange          orange
; 0011          lime green      peach
; 0100          orange          pink
; 0101          blue green      purple
; 0110          pink red        dark blue
; 0111          aqua            blue
; 1000          purple          blue
; 1001          aqua            blue
; 1010          purple          aqua
; 1011          blue            green/blue
; 1100          blue purple     melon green
; 1101          dk.blue         olive
; 1110          grey            lt.green
; 1111          grey            orangy



SPHunter
        .word SSHunter          ; stand
        .word SSHunter          ; jump

SPPOOF  .word SSPOOF            ; stand

SPTwoShoes
        .word SSTwoShoes        ; stand
        .word SSTwoShoes        ; jump

HumanSprite
        .word SS_Normal          ; normal
        .word SS_Jump            ; jump
        .word SS_Sad
        .word SS_Blink

JumpVector      .byte 5,4,3,2,1,-1,-2,-3,-4,-5





SSHunter


        .byte 12


        .byte ORANGE + %1001

        .byte %00100000
        .byte %10110000  
        .byte %01011000  
        .byte %11111100  
        .byte %01111110  
        .byte %10101010  
        .byte %01111111  
        .byte %11110011  
        .byte %11101101  
        .byte %01111110 

        .byte BLANK_LINE

        .byte RED + %1000

        .byte %10010000
        .byte %01011000  
        .byte %10111100  
        .byte %00111100  
        .byte %11111110  
        .byte %01101010  
        .byte %10111111  
        .byte %11110011  
        .byte %11101101  
        .byte %01111110 

        .byte BLANK_LINE

SSTwoShoes

        ; rabbit

        .byte 12

        IF COMPILE_VERSION = PAL
                .byte %00101000
        ELSE
                .byte %00101010
        ENDIF

        .byte %00111100 ;   ****@
        .byte %00111100 ;   ****@
        .byte %00001110 ;  ***@
        .byte %01110010 ;  *@@***
        .byte %11111111 ; ********
        .byte %11111101 ; * ******
        .byte %11111111 ; *******#
        .byte %01111111 ; #*******
        .byte %01100100 ;   *  **
        .byte %00110010 ;  * @**

        .byte BLANK_LINE

        IF COMPILE_VERSION = PAL
                .byte %00100111               ; greenish
        ELSE
                .byte %00101001
        ENDIF
        
        .byte %01111100
        .byte %01111100
        .byte %00111110
        .byte %01111110
        .byte %11111111
        .byte %11111101
        .byte %01111111
        .byte %01111110
        .byte %01100100
        .byte %00111010

SSLurkEgg1

        .byte BLANK_LINE

                ; merged...

;        .byte 0
        .byte %00101111
        .byte %00000000  
        .byte %00000000  
        .byte %00000000  
        .byte %00000000  
        .byte %00000000  
        .byte %00011000  
        .byte %00111100  
        .byte %00111100  
        .byte %00111100  
        .byte %00011000  

SSPoints_1UP

        .byte BLANK_LINE


                ; merged...

        .byte %01111100

        .byte %00011110
        .byte %00110101
        .byte %01101111
        .byte %01110001
        .byte %00111110
        .byte %00100100
        .byte %00110110
        .byte %00000000                 ; actual data
        .byte %00000000
        .byte %00000000

SSPoints_100

        .byte 0 ;12

                ;merged

        .byte %01111100

        .byte %01001010 ;  *  * *
        .byte %11010101 ; ** * * *
        .byte %01010101 ;  * * * *
        .byte %01001010 ;  *  * *
        .byte %01000000 ;  *
        .byte %01000000 ;  *
        .byte %00000000
        .byte %00000000
        .byte %00000000
        .byte %00000000

SSPoints_300


        .byte BLANK_LINE

                ; merged...

        .byte %01111100

        .byte %10001010 ; *   * *
        .byte %01010101 ;  * * * *
        .byte %11010101 ; ** * * *
        .byte %00101010 ;   * * * 
        .byte %00100000 ;   *
        .byte %11000000 ; **  
        .byte %00000000
        .byte %00000000
        .byte %00000000
        .byte %00000000

SSPoints_500

        .byte BLANK_LINE

                ; merged...

        .byte %01111100

        .byte %11001010 ; **  * *
        .byte %10010101 ; *  * * *
        .byte %11010101 ; ** * * *
        .byte %00101010 ;   * * *
        .byte %00100000 ;   *
        .byte %11000000 ; **
        .byte %00000000

Div5                            ; merged sprite/table ... do not move!
        .byte %00000000
        .byte %00000000
        .byte %00000000
        .byte BLANK_LINE


div set 1
        REPEAT 3
        REPEAT CUBEXY
        .byte div
        REPEND
div set div+1
        REPEND

Mod5

        REPEAT 4
div set 0
        REPEAT CUBEXY
        .byte div
div set div+1
        REPEND
        REPEND


;------------------------------------------------------------------------------
        IF TITLESCREEN

TitleScreen

        lda #0
        sta TS_line
        sta COLUBK


TitleScreen2


        jsr VBPreliminary

                ; "2001"
        ldx #0
        stx S0
        stx S1
        inx
        stx S3
        stx S4
        inx
        stx S5
        inx
        stx S2


        lda TS_line
        lsr
        lsr
        cmp #8
        bcc l92
        lda #8
l92     sta TS_width


                ; << lots of time in here to do stuff
                ; (approx 2800 cycles)

        jsr VBFinal

                ; Center the title vertically

        IF COMPILE_VERSION = PAL
                ldy #33
        ELSE
                ldy #32
        ENDIF

Bline   sta WSYNC
        dey
        bne Bline

        IF COMPILE_VERSION = PAL
                lda #255                 ;
        ELSE
                lda #238                 ; NTSC
        ENDIF
        
        sta TIM64T


        ldx #15

Title   lda TS_width
        sta TS_temp

ALine   sta WSYNC

        lda #0
        sta PF0        
        lda QbTitle-1,x
        sta PF1
        lda QbTitle1-1,x
        sta PF2

        iny
        sty COLUPF
        nop

        lda QbTitle2-1,x
        sta PF0
        lda QbTitle3-1,x
        sta PF1
        lda QbTitle2-1,x
        and #$0f
        sta PF2

        dec TS_temp
        bpl ALine

        dex
        bpl Title

        lda #%00001000
        jsr ScoreLine

ThankES2
        lda INTIM
        bne ThankES2

        inc TS_line
        beq GoWarm

        jsr Interact
        beq TitleScreen2                           ; user intervention for restart

GoWarm  rts
        ENDIF

;------------------------------------------------------------------------------

KillPlayer

        lda Cubes
        and #CUBES_DEMO_MODE
        bne WarmBoot                            ; don't worry about stuffed-up stack :)

                ; Player loses a life

        lda #YELLOW + %10
        sta BackColour

        jsr DisplayLivesScore
        jsr TieToCube

        lda #TYPE_CORPSE
        jmp ChangeCreature                      ; JSR/RTS optimisation

;------------------------------------------------------------------------------

GetPattern

        lda Cubes
        asl
        bcs FT2a                        ; CC when NOT completed!

        lda #1
        cmp timer+1
FT2a    lda #8
        tay
        bcc FTargA                      ; not time... normal display instead
        and timer
FTargA
        asl                             ; 0 or 16
        rts


;------------------------------------------------------------------------------

TargetBorder

        jsr GetPattern
        beq Zbord
        lda #31
Zbord   sta MTA_31
        sta MTA_31+1
        rts

;------------------------------------------------------------------------------


overscan

        sta WSYNC

;        IF COMPILE_VERSION = PAL
;        lda #61                 ; 30+ scan lines of overscan
;        ELSE
;        lda #46                 ; NTSC
;        ENDIF
        
;        sta TIM64T
        
        lda timer
        lsr
        bcs MCLimit2             ; limit speed of *ALL* cube movement

        jsr MoveCube
MCLimit2



        ; Perform AI/logic for creatures
  
        ldx #0
        jsr DoAI


                ; << lots of time in here to do stuff
                ; (approx 2800 cycles)



        lda Cubes
        bmi Finished                    ; halt when screen complete

        ldx #1
        jsr DoAI

Finished




ThankES lda INTIM
        bne ThankES

        IF COMPILE_VERSION = PAL
                sta WSYNC
                sta WSYNC
        ENDIF

        beq verticalblank


;--------
; start
; program entry point


start

;        lda #$6D
;        sta rnd
;        sta rnd+1

WarmBoot

        sei
        cld

        ldx #$FF
        txs

        ;--------
        ; one-time initialization
        ; happens at system reset, no other time

        inx
        stx SWACNT

        lda rnd+1
;        and #31
        sta level

        jsr NextLevel

        lda Cubes
        ora #CUBES_DEMO_MODE
        sta Cubes

        lda SerialNumber
        sta PlayerScore

        IF TITLESCREEN
                jsr TitleScreen
                beq verticalblank
                jsr BeginGame
        ENDIF
        

verticalblank
VBPreliminary

        ; writes the vertical synch signal for the start of a frame
        ; sets up timer so we can keep track of available time - note
        ; there are 37 lines available before the screen needs to start
        ; drawing - approx 2800 cycles therefore available for game
        ; calculations


        lda #D1+D6
        sta VBLANK              ; start with tia disabled


    ; typical frame consists of 3 vertical sync (VSYNC) lines,
    ; 37 vertical blank (vblank) lines,
    ; 192 tv picture lines
    ; 30 overscan lines

    ; ref: nick bensema's wonderful "how to draw a playfield"

    ; require 3 VSYNC lines at start of frame
    ; ref: spg pp.1

        sta WSYNC
        sta VSYNC               ; D1 in A

        sta WSYNC
        sta WSYNC
        sta WSYNC

        ldy #3
Synch   sta WSYNC
        dey
        bne Synch

        sty VSYNC

    ; there's 76 cycles available on each of these lines
    ; three required for the VSYNC - don't cut it too close
    ; but do interleave other code here

;        sta WSYNC               ; 1st
;        sta WSYNC               ; 2nd   TV SYNCH (2 tested OK on PAL, 1 FAILED)
;        sta WSYNC               ; 3rd, to satisfy ES :)


    ; we have just 37 (more) lines (@2812 cycles) of vblank, so we have to
    ; make sure the screen draw doesn't get stomped on.  the timer will tell
    ; us if we've screwed up.  timer done just before the last WSYNC to
    ; give maximum possible cycle count availability/accuracy

        IF COMPILE_VERSION = PAL
        lda #72                 ; PAL
        ELSE
        lda #31                 ; NTSC
        ENDIF

        sta TIM64T              ; 40/48 scan lines -2 (above)


;        lda #0
;        sta VSYNC               ; stop vertical synch


    ; screen is now scanning, but tia suppressed
    ; begins again when 0 -> vblank

    ; this is running after the TV VSYNC has been sent, but before
    ; we need to start drawing lines.  there are 37 lines, just around
    ; 2800 cycles.  gotta be sure sure that we don't overflow that time
    ; otherwise we'll get screen jitter

;------------------------------------------------------------------------------


        lda BackColour
        sta COLUBK

        lda Cubes
        and #D6
        tax

        lda level
        asl
        asl
        asl
        asl
        bne nDark                               ; avoid the 'dark' grid which doesn't look good
        lda #10*16
nDark



        IF COMPILE_VERSION = PAL
                ora #%0100
        ELSE
                ora #%0010
        ENDIF

        IF ATTRACT_MODE
                
                dex
                bmi EOL                 ; only do it in demo mode
                eor BackColour          ; gives demo effect
EOL
        ENDIF
        
        sta COLUPF


        dec timer                       ; one tick

        jsr TargetBorder                ; fix up MTA_31 (now needed due to alternating draw)


;-----

                ; To reduce total time usage, the draw of the main board and the target board
                ; are alternated.  This should fix the overusage of time :))



        lda timer
        lsr
        bcc CTabloo
        
        jmp DoDraw
CTabloo



                ; When time runs out, generate new tableaux
                ; This is also the first-time init of the tableaux

        lda Cubes
        bmi nHTimer                     ; in 'level complete' mode

        lda PlayerCube
        bpl noPenalty
        lda #$97
        jsr DecScore                    ; penalise player for slowness

        lda PlayerScore
        ora PlayerScore+1
        ora PlayerScore+2
        bne noPenalty

        jsr ShootPlayer

noPenalty

        lda timer
        bne nHTimer
        dec timer+1
        bpl nHTimer

        inc timer+1
        bne ResetTableaux               ; abnormal wrap (1st time init)

                ; so, run out of time!

        lda Cubes
        and #CUBES_DEMO_MODE
        beq nDemoM
        jmp WarmBoot                    ; out of time aborts demo mode
nDemoM

        asl PlayerCube
        sec
        ror PlayerCube                  ; set the penalty flag

                ; Should be here when first starting a level (ie: timer init'd to 0 on level)

ResetTableaux

                ; This resets and generates a random pattern on the tableaux
                ; Calculate a graduated time for completing the screen

        sec
        lda #31
        sbc level
        and #%11111
        lsr
        lsr
        sta timer+1

        jsr GetCubeTotal
        sta CT_Cubes
 
        ldx #3
ClGrid2 lda Grid,x
        and #%01010101                  ; clear all (previous) home positions
        sta Grid,x
        dex
        bpl ClGrid2

CTab
                ; Select a random placement for a cube
                ; returns X,Y position
                ; DON'T MAKE THIS RANDOM/RANDOM, as it will delay/jiggle the screen

        jsr Random0to3
        lda rnd+1
        and #3
        tax

CheckT  lda Grid,x
        and TargetBit,y
        beq FTab

        lda NextX,x
        tax
        bne CheckT
        lda NextX,y
        tay
        bpl CheckT

FTab    lda Grid,x
        ora TargetBit,y
        sta Grid,x

        dec CT_Cubes
        bne CTab

        jmp noCheck


                ; end of ResetTableaux
nHTimer


CheckTableaux

                ; Flash the border of the target area if we're running out of time
                ; (OR if screen complete)

        jsr TargetBorder
        and #16                         ; caters for '31' -> 16
DClear  sta TargetDisplay-1,y
        dey
        bne DClear

                ; CheckTableaux

        lda #15
        sta CT_index

CTIter  lda CT_index
        and #3
        tay
        lda CT_index
        lsr
        lsr
        tax

        lda Grid,y
        and TargetOwnerBit,x
        cmp TargetOwnerBit,x
        beq Correct             ; correct square, correct position

        cmp TargetBit,x
        bne Flasher             ; no target

                ; So there's a target, but no cube.  It has to flash

        lda timer
        and #8
        beq Flasher             ; toggle controls flash rate

Correct tya
        asl
        tay

                ; Turn on any display for this square

        lda TargetDisplayBitL,x
        ora TargetDisplay,y
        sta TargetDisplay,y

        lda TargetDisplayBitR,x
        ora TargetDisplay+1,y
        sta TargetDisplay+1,y

Flasher dec CT_index
        bpl CTIter
        bmi noCheck

                ; end of CheckTableaux (X=0, Y=0)


DoDraw


DrawErase

                ; Flash cubes (by skipping draw) when the tableaux has been completed

        lda Cubes
        bpl NDraw                       ; not complete
        lda timer
        and #8
        beq SkipDrE                     ; skip draw on tableaux completion
NDraw



        ldx #NUMBEROFCUBES-1
DrawC2

        lda CubeX,x
        bmi NextDr

        ; Draw/Erase a single cube at x,y location

        ; pass
        ; X     = index of cube (preserved)
        ;       CUBE MUST EXIST (That is, no passing empty slots)

        ; the y position simply gives us the ROW (*5) of the start of draw
        ; the X position is mangled to give us byte information.
        ; TRICKY due to the odd layout of playfields

        ; use indexing into (predrawn) table lines to make it simple


        ldy CubeX,x
        cpy #16
        bcc NoBottom

                ; Draw the sliding cube on the bottom

        lda PreDrawn3,y
        sta TargetDisplay+8
        lda PreDrawn4,y
        sta TargetDisplay+9            ; this is pretty neato... just two bytes for the bottom sliding cube :))

NoBottom

        stx DC_saveX

                ; Draw in the grid area playfield
        
        lda CubeYDir,x
        eor #15
        and #15
        tax                             ; playfield location of start

        lda #3
        sta DC_CubeLine

DrawLoop

        lda PreDrawn0,y
        ora playfield,x
        sta playfield,x

        lda PreDrawn1,y
        ora playfield+PFC,x
        sta playfield+PFC,x

        lda PreDrawn2,y
        ora playfield+PFC+PFC,x
        sta playfield+PFC+PFC,x

        inx

        dec DC_CubeLine
        bpl DrawLoop

        ldx DC_saveX

NextDr  dex
        bpl DrawC2


SkipDrE 
noCheck

        jsr SquirtCube                  ; Attempt to feed a cube bottom-right position


;------------------------------------------------------------------------------

VBFinal


endvb   lda INTIM
        bne endvb

        sta WSYNC
        sta VBLANK              ; end vblank period
;        sta WSYNC

        IF COMPILE_VERSION=PAL
                lda #19
        ELSE
                lda #18
        ENDIF
        sta T1024T



                ; fall through
;-----------------^^^^^^^^^^^^-------------------------------------------------

DrawScreen

    ; the kernel
    ; line by line we have to hold hands with the tia, passing it data
    ; so that it draws the screen for us.
    
    ; wait for vertical blank period to end
    ; timer counts down 1 every 64 cycles, so just wait for it
    ; it may already have expired


;        lda #%11111100
;        sta COLUP0
;        sta COLUP1

;        lda #233
;        sta TIM64T                     ; 200.4 scan lines

;        sta WSYNC                       ; not 100% required
;        sta WSYNC
;        sta WSYNC

;------------------------------------------------------------------------------


SpriteHandling

        ; Sets up the player sprites - called at the top of screen just before sprites
        ; If there's a score on display, its after the score


        ; Setup vars for graphics draw of sprites
        ; HORRIBLY INEFFICIENT-LOOKING

        ; setup (sprite+x) to account for new draw system
        ; must point n bytes BEFORE definition, and still not cross page boundary (sheesh)
        ; also setup SpriteEnd pointers appropriately


        ldx #1

SetupSprite

                ; Mirror the sprite IFF the object is/has been facing left
                ; The mirroring is currently cleared for digit display in ScoreLine

        lda PlayerFlags,x
        lsr                             ; mirror -> D3 -> hardware
        sta REFP0,x                     ; set mirror appropriately

        lda PlayerCube,x
        and #MASK_TYPE
        lsr
        lsr
        tay

        lda SpriteVector,y
        sta DS_Sprite
        lda SpriteVector+1,y
        sta DS_Sprite+1                 ; --> vector to p2

        lda PlayerFlags,x
        and #MASK_FRAME
        tay
        lda (DS_Sprite),y               ;--> table of animation pointers
        pha
        iny
        lda (DS_Sprite),y
        sta DS_Sprite+1
        pla
        sta DS_Sprite                   ; --> frame of animation

        cpx #0
        bne nSP                         ; don't save player data
        pha
        lda DS_Sprite+1
        pha
nSP


                ; Frame block data consists of
                ; +0                    D6-D0 byte offset to multiplex pattern - 1
                ;                             points to the start of the sprite frame (colour)
                ;                       D7    this frame has "rising" capability

                ; --> single frame itself consists of...
                ; +0                    colour of frame
                ; +1-9                  (OPTIONAL, see D7 above) 0-bytes allow "rising" capability
                ; +10 (or 1)            BLANK_LINE (ie: 0) required by sprite system
                ; +11-20 (or 2-11)      10 lines of sprite data


        lda timer
        lsr                             ; D0 --> C

        ldy #0
        lda (DS_Sprite),y
        bcc NoMulti
        tay                             ; offset to multiplex pattern #2
NoMulti

        iny                             ; --> colour
        lda (DS_Sprite),y
        sta COLUP0,x                      

        sty DS_temp                     ; --> start of the 10 bytes of frame data

        txa
        asl
        tay

        sec                             ; yes, +1
        lda DS_temp
        adc DS_Sprite
        adc #SPRITE_HEIGHT
        sta Sprite,y
        lda DS_Sprite+1
        adc #0
        sta Sprite+1,y

                ; Have finished calculating what shape to display, now WHERE (vertically) to display it?

        lda PlayerCube,x
        and #MASK_CUBE
        tay

        lda CubeYDir,y
        and #15                         ; cube Y coord
        sta DS_temp
        asl
        adc DS_temp                     ; 3y (3 sprite lines/1 y for cubes)
        sec
        sbc PlayerY,x

        adc #SPRITE_Y_ADJUST
        sta SpriteEnd,x                 ; double-scanline of end of sprite
     
        txa
        asl
        tay

        sec
        lda Sprite,y
        sbc SpriteEnd,x
        sta Sprite,y
        lda Sprite+1,y
        sbc #0
        sta Sprite+1,y

        dex
        bpl SetupSprite


                ; The colour address must be offset in the same way as the player address


        pla
        sta DS_temp+1
        pla
        sta DS_temp
        
        sec
        lda Sprite
        sbc DS_temp
        sta DS_temp
        lda Sprite+1
        sbc DS_temp+1
        sta DS_temp+1                   ; calculated offset of P0 sprite data

        clc
        lda #<PlayerColour
        adc DS_temp
        sta ColourPlayer
        lda #>PlayerColour
        adc DS_temp+1
        sta ColourPlayer+1


PositionSprites

        ; Set the horizontal position of the two sprites, based upon their coordinates
        ; This uses the tricky 2600 method of positioning (RESPx, HMPx, etc)

        ; Algorithm invented a looong time ago by persons unknown
        ; Re-invented Feb2001 by yours truly, then optimised according to code by Thomas Jenztsch

        sta WSYNC
        sta HMCLR                       ; clear any previous movement

        ldx #2                          ; sprite index
PosSP   
        lda PlayerX-1,x

        tay

        lsr
        lsr
        lsr
        lsr
        
        sta PS_temp

        tya
        and #15

        clc
        adc PS_temp

        ldy PS_temp

        cmp #15
        bcc NH
        sbc #15
        iny
NH

                ; Use remainder for fine adjustment
        eor #7
        asl
        asl
        asl
        asl

        sta HMP0-1,x                      ; fine movement
        sta WSYNC

        jsr Ret
        bit 0                           ; 15 cycles = 3 loops :)

Jiggle  dey
        bpl Jiggle

        sta RESP0-1,x

        dex
        bne PosSP

        sta WSYNC
        sta HMOVE
        sta WSYNC

        stx VDELP0
        stx CTRLPF
        stx NUSIZ0
        stx NUSIZ1

; === end of sprite handling


        iny
        sty MTA_ZERO
        sty MTA_ZERO+1

        jsr MB4                 ; (x=0)
        jsr MB4
        jsr MB4

;------------------------------------------------------------------------------


    ;----
    ; cycle thru drawing line by line...
    ; time is precious here... there are just 228/3 cycles per line
    ; (giving us just 76 cycles) and that's not counting the WSYNC,
    ; should you be writing that. There are 68 pixels (/3 = cycles)
    ; in the hblank period, which starts at the conceptual beginning
    ; of a line on this machine

;        lda #0                 ; GUARANTEED A=0 (above)
        sta MT_TargetMask       ; mask for border of target area

        jsr MB4a

        ldx #18                          ; << character line
        jmp MTargetArea                 ; the whole shebang!



;------------------------------------------------------------------------------

Owned   ldx CM_CubeIndex
        ldy CM_Direction

        sec                     ; fail
        rts


CheckMove

        ; Determine if it is possible to move in a desired direction
        ; if so, current position is vacated and new position claimed


        ; y = direction (guaranteed 0-3, preserved)
        ; x = cube index (preserve!)  MUST BE A LEGIT CUBE, NOT JUST AN INDEX
        ; AI_Player = player# (could be FF)


        ; returns carry C = success
        ; S = fail
        ; X,Y preserved

        stx CM_CubeIndex
        sty CM_Direction
                
      
         ; apply movement

        lda CubeX,x
;        and #31                OPTIMISED, ASSUMING OTHER CUBEX BITS (except D7) UNUSUED
        cmp #16
        bcs Owned               ; offscreen - no moving
        tax

        lda DirMoveX,y
        adc Div5,x              ; now a horizontal index (0-3)  (RELY ON CARRY C)
        cmp #4                  ; an illegal position?
        bcs Owned               ; can't move there

        sta CM_Column

        ldx CM_CubeIndex
        lda CubeYDir,x
        and #15
        tax

        lda DirMoveY,y
        adc Div5,x              ; (RELY ON CARRY C)
        cmp #4                  ; an illegal position?
        bcs Owned               ; can't move there
        
        tay
        ldx CM_Column


        lda Grid,y              ; row of grid
        and OwnershipBit,x
        bne Owned               ; can't go there!


        jsr GrabOwnership               ; OK, so the grid is available... grab it!

                ; Score for player moving cube to correct position

        and TargetBit,x
        beq NoGoal

        lda PlayerCube
        eor CM_CubeIndex
        and #MASK_CUBE
        bne NoGoal                      ; not player's cube

        ldy #0
        lda #$10
        jsr AddSmallScore
        
NoGoal

        ldy #SOUND_SLIDING
        jsr Sound               ; sliding

                ; free the grid currently occupied

        ldx CM_CubeIndex
        jsr GetGridIndex

        lda OwnershipBit,x
        eor #$FF
        and Grid,y
        sta Grid,y

                ; Merge new direction into cube information

        ldy CM_Direction
        ldx CM_CubeIndex
        lda CubeYDir,x
        and #%00001111                  ; keep all but direction bits
        ora DirShifted,y
        sta CubeYDir,x

        clc
        rts


;------------------------------------------------------------------------------

GrabOwnership

        ; Set ownership bit in grid
        ; pass y = grid row, x = grid column

        lda Grid,y
        ora OwnershipBit,x
        sta Grid,y
        rts

;------------------------------------------------------------------------------

CheckJump
; Optimized 14/2

        ; pass X = cube index
        ; Y = direction (only use 0-3 - out of range NOT checked)
        ; AI_Player = index of player doing the jump
        ; Tries to jump creature in desired direction
        ; returns: x,y buggered
        ; carry C = jump taken, S = did not jump

        sty CJ_Direction

        lda CubeX,x
        bmi NoJump              ; dead cube (!)
        lda CubeYDir,x
        bpl NoJump              ; my cube is moving
        and #15

        clc
        adc DirMove5Y,y         ;  now a vertical coordinate
        cmp #16                 ; an illegal position?
        bcs NoJump              ; can't move there

        sta CJ_NewY

        lda CubeX,x
;        and #31                OPTIMISED, ASSUMING OTHER CUBEX BITS (except D7) UNUSUED
;        cmp #16
;        bcs NoJump              ; feeding - not a valid source (already catered, above)

        adc DirMove5X,y         ; CC assumed... now a horizontal coordinate
        cmp #16                 ; an illegal position?
        bcs NoJump              ; can't move there

        sta CJ_NewX


                ; Find the cube occupying that grid position (if any)

        ldx #NUMBEROFCUBES-1

FindTarget

        lda CubeX,x
        cmp CJ_NewX
        bne NextTarget
        
        lda CubeYDir,x
        bpl NextTarget          ; potential destination is moving - fail
        and #15
        cmp CJ_NewY
        beq GotTarget           ; succeeds on X,Y

NextTarget

        dex
        bpl FindTarget
        rts

GotTarget
              
        stx CJ_CubeNumber
 
        ldx AI_Player
        bmi NoJump                      ; virtual

        lda PlayerCube,x
        and #255-MASK_CUBE
        ora CJ_CubeNumber
        sta PlayerCube,x                ; record new cube for player

        lda #MODE_JUMP + 10
        sta PlayerCounter,x

        ldy CJ_Direction
        lda AdjustY,y
        sta PlayerY,x

                ; store the jump direction

        lda PlayerFlags,x
        and #255-MASK_DIRECTION
        lsr
        ora DirShifted,y
        rol
        sta PlayerFlags,x

        clc
        rts

NoJump

        sec
        rts



;------------------------------------------------------------------------------


Random
        ; Generate a pseudo-random number

       lda rnd
       eor timer
       lsr
       lsr
       sbc rnd
       lsr
       ror rnd+1
       ror rnd
       ror rnd
       lda rnd
       rts

Random0to3

        jsr Random
        and #3
        tay
        rts

;------------------------------------------------------------------------------


TargetDisplayBitL

        .byte %00001100
        .byte %00000011

;\--> merges with next table... don't move 'em!        

TargetDisplayBitR

        .byte %00000000
        .byte %00000000
        .byte %00000011
        .byte %00001100


;------------------------------------------------------------------------------

IsCorrectPlace

        ; CANNOT USE OVERLAY VARS

        ; Check target grid to see if a cube belongs here
        ; pass x = cube index
        ; returns carry S = yes, C = no
        ; x,y buggered

                ; fall through
;-----------------^^^^^^^^^^^^-------------------------------------------------

GetGridIndex

        ; CANNOT USE OVERLAY VARS

        ; Return an x,y index to the grid       (does a bit of exta work to cater for IsCorrectPlace call)
        ; pass x = cube #
        ; returns x = grid column index, y = grid row index
        ; trims out-of-bounds x to board right-side (ie: 3)

        lda CubeYDir,x
        and #15
        tay
        lda Div5,y
        tay

        lda CubeX,x
;        and #31                OPTIMISED, ASSUMING OTHER CUBEX BITS (except D7) UNUSUED                 ; GUARANTEED <=15?
        cmp #16
        bcc GL16
        lda #15
GL16    tax
        lda Div5,x
        tax

                ; fall through
;-----------------^^^^^^^^^^^^-------------------------------------------------

CheckTargetOwner

        ; CANNOT USE OVERLAY VARS

        ; Check grid location for correctness
        ; pass y = grid row, x = grid column
        ; returns BEQ (BCS) correct
        ;       BNE (BCC) incorrect

        lda Grid,y
        and TargetOwnerBit,x
        cmp TargetOwnerBit,x
        rts

;------------------------------------------------------------------------------


MB4a    ldx #3                          ; GUARANTEED RETURN A=0
MB4     jsr MBlock                      ; "
MBlock3 jsr MBlock
        jsr MBlock

                ; fall through
;-----------------^^^^^^^^^^^^-------------------------------------------------

MBlock

        ; GUARANTEED RETURN: A=0
        ; preserve y (+1)

     ; LINE 1 -------------------------------

        sta WSYNC               ; 3

        lda #0                  ; 2
        sta PF0                 ; 3 ->8              (@5)
        sta PF1                 ; 3 ->11        < 28 (@8)
        sta PF2                 ; 3 ->14        < 39 (@11)

;\\22
        jsr Ret                 ; 12
        pha
        pla
        bit 0
;//

        sec                     ; 2

        ;--------------------------
        ; SPRITE DRAW

        tya                     ; 2
        sbc SpriteEnd           ; 3
        adc #SPRITE_HEIGHT      ; 2
        bcs .MBDraw3            ; 2(3)

        nop                     ; 2
        nop                     ; 2
        sec                     ; 2
        bcs .skipMBDraw3          ; 3

.MBDraw3
        lda (Sprite),y          ; 5
        sta PLAYER0SHAPE        ; 3

.skipMBDraw3

        ;-->18 cycles used
        ;--------------------------

        lda (ColourPlayer),y    ; 5
        sta COLUP0              ; 3

;->64

     ; LINE 2 -------------------------------

        ;sta WSYNC              ; 3

        lda BlockIndex,x        ; 4
        sta PF0                 ; 3
        lda BlockIndex+1,x      ; 4
        sta PF1                 ; 3 ->17        < 28 (@14)
        sta PF2                 ; 3 ->20        < 39 (@17)



    ; right pf




        ;--------------------------
        ; SPRITE DRAW

        tya                     ; 2
        sbc SpriteEnd+1         ; 3
        adc #SPRITE_HEIGHT      ; 2
        bcs .MBDraw2            ; 2(3)

        nop                     ; 2
        nop                     ; 2
        sec                     ; 2
        bcs .skipMBDraw2        ; 3

.MBDraw2
        lda (Sprite+2),y        ; 5
        sta PLAYER1SHAPE        ; 3

.skipMBDraw2

;\\2
        bit 0
;//


        lda (ColourPlayer),y    ; 5

        iny                     ; 2

        sta COLUP0              ; 3

        ;-->30 cycles used
        ;--------------------------

        lda BlockIndex+2,x      ; 4
        sta PF0                 ; 3 ->41        < 49 (@37)

        lda #0                  ; 2
        sta PF1                 ; 3
        sta PF2                 ; 3 ->54[5]     < 66 (@49[50])

                ; GUARANTEED RETURN : A=0

        rts                     ; 6

;------------------------------------------------------------------------------

FeedCube

        ; Feed a new cube onto the grid
        ; Count the number of cubes actually IN ACTION.
        ; If there's more to come on, AND the bottom right square is free...
        ; FEED on a cube

        ; pass Y = row (0-15), X = column (0-15) or 31 for special feeder
        ; returns X = cube index of fed cube (or $FF if none)


        stx FC_CubeX
        sty FC_CubeY


        lda Cubes
        and #7
        tax
        dex
        stx FC_BlankSpot
        bmi NotVacant                   ; already all active


;        cpx #NUMBEROFCUBES-2
;        bcs NoPend
;        lda CubeX+1,x
;        cmp #24
;        bcs NoPend


                ; There's still cubes to come... possibly show a pending cube at right edge
                ; See if the previous cube is fully fed, first...

        ldx #NUMBEROFCUBES-1
FSlider lda CubeX,x
        bmi nValid                      ; dead cube
        cmp #24
        bcs NoPend
nValid  dex
        bpl FSlider

        lda #%01111000
        sta TargetDisplay+9             ; show the static 'pending' cube

NoPend




        lda FC_CubeX
        cmp #16
        bcc NotSlider
        lda #15
NotSlider
        lsr
        lsr
        tax

        tya
        lsr
        lsr
        tay

        lda Grid,y
        and OwnershipBit,x
        bne NotVacant                   ; can't feed: square occupied!

 
                ; Place cube onscreen.  Could be anywhere.

        ldx FC_BlankSpot                ; last blank spot

        lda FC_CubeX
        sta CubeX,x
 
        lda FC_CubeY
        ora #$80                        ; stationary (as it could be put anywhere, not just 'feeding')
        sta CubeYDir,x 

        ldy #SOUND_SLIDING
        jsr Sound

        jsr GetGridIndex
        jsr GrabOwnership

        dec Cubes                       ; one less to do

NotVacant

        ldx FC_BlankSpot
        rts



;------------------------------------------------------------------------------

SquirtCube
                ; Insert a new cube feeding bottom right

        ldx #31
        ldy #15
        jsr FeedCube
        bmi noJuice

        lda #%00111111
        sta CubeYDir,x                  ; moving left, bottom of screen

noJuice
        rts

;------------------------------------------------------------------------------


joystickdir
        .byte $FF,$FF,$FF,$FF,$FF,$FF,$FF,1,$FF,$FF,$FF,3,$FF,2,0,$FF

;------------------------------------------------------------------------------


; Types: uses D3-D6 of 'PlayerCube'

TYPE_PLAYER     = %00000000             ; 0
TYPE_LURKER     = %00001000             ; 1
TYPE_FRUIT      = %00010000             ; 2
TYPE_POINTS     = %00011000             ; 3
TYPE_SHIFTER    = %00100000             ; 4
TYPE_2SHOES     = %00101000             ; 5
TYPE_HUNTER     = %00110000             ; 6
TYPE_CORPSE     = %00111000             ; 7             dead player (unresponsive)
TYPE_POOF       = %01000000             ; 8             creature disappearing


TypeVector      .word TypePlayer
                .word TypeLurker
                .word TypeFruit
                .word TypePoints
                .word TypeShifter
                .word Type2Shoes
                .word TypeHunter
                .word TypeCorpse
                .word TypePOOF

TypePlayer      .word ShifterInit-1
                .word AINormal-1                        ; MODE_PLAYER_NORMAL
                .word 0
                .word AIJump-1                          ; MODE_PLAYER_JUMP
                .word 0 ;AILand-1                          ; MODE_LAND
                .word AILock-1                          ; MODE_PLAYER_LOCK
                .word KillPlayer-1                      ; MODE_PLAYER_KILL

TypeCorpse      .word ShifterInit-1
                .word CorpseNormal-1
                .word 0         ; toss
                .word 0         ; jump
                .word 0         ; land
                .word 0         ; lock  << ?
                .word 0         ; kill
                .word CorpseDeathwatch -1        ; kill


TypeLurker      .word ShifterInit-1
                .word LurkNormal-1
                .word 0                                 ; cannot toss!
                .word LurkHatch-1

TypeFruit       .word FruitInit-1
                .word FruitNormal-1
                .word 0                                 ; fruit doesn't get tossed

TypePoints      .word PointsInit-1
                .word PointsNormal-1
                .word 0                                 ; points don't get tossed

TypeShifter     .word ShifterInit-1
                .word ShifterNormal-1
                .word AITossOff-1
                .word AIJump-1

Type2Shoes      .word ShifterInit-1                     ; re-use
                .word TwoShoesNormal-1
                .word AITossOff-1
                .word AIJump-1

TypeHunter      .word ShifterInit-1
                .word HunterNormal-1
                .word AITossOff-1
                .word AIJump-1

TypePOOF        .word ShifterInit-1
                .word POOFNormal-1
                .word 0                                 ; non tossable

;------------------------------------------------------------------------------

GetAIVector

        ; Returns a pointer to the creature's AI vector table
        ; use the creature mode to index to this

        ; pass  x = creature


        lda PlayerCube,x
        and #MASK_TYPE
        lsr
        lsr
        tay                                     ; creature type, word index

        lda TypeVector,y
        sta AI_Vector
        lda TypeVector+1,y
        sta AI_Vector+1

        rts

;------------------------------------------------------------------------------

GetModeIndex

        ; Returns the index for the mode passed in a
        ; adds 1, so pointer is pointing to high byte of word tables
        ; pass  a = mode

        lsr
        lsr
        lsr
        lsr
        and #%1110
        tay                                     ; creature mode, word index
        iny
        rts


;------------------------------------------------------------------------------

AIJump

        ldy #SOUND_JUMP
        jsr Sound

        jsr Weapon

        lda #%0
        jsr CountDown                   ; -->y
        bcc EndJump

        lda #FRAME_JUMP
        jsr SetFrame

        tya
        and #3
        sta AIJ_YSlower

;        sec                             ; CS!
        lda PlayerY,x
        sbc JumpVector,y
        sta PlayerY,x

        lda PlayerFlags,x
        asl
        rol
        rol
        rol
        and #3
        tay                             ; direction movement

        lda DirMoveX,y
        asl
        clc
        adc PlayerX,x
        sta PlayerX,x
         
        lda AIJ_YSlower
        beq nY
            
        sec
        lda PlayerY,x
        sbc DirMoveY,y
        sec
        sbc DirMoveY,y
        sta PlayerY,x

nY      rts


;------------------------------------------------------------------------------

EndJump lda #MODE_NORMAL
        jsr ChangeMode

        lda #FRAME_STAND
        jsr SetFrame

        ldy #SOUND_OFF
        jsr Sound

                ; fall through
;-----------------^^^^^^^^^^^^-------------------------------------------------

TieToCube

        ; Horizontally position a sprite to where a cube is
        ; Screwy sprite positioning by the hardware makes this a bit tricky

        ; X = sprite object to tie (0 or 1) (preserved)

        lda PlayerCube,x
        and #MASK_CUBE
        tay

        lda CubeX,y
        asl
        asl
        adc #SPRITE_X_ADJUST
        sta PlayerX,x

        lda #0
        sta PlayerY,x

        rts

;------------------------------------------------------------------------------


Sound
        ; Sounds are mini-programs
        ; there are just two sound channels, each has a program which it runs
        ; the program consists of a volume, frequency, control, duration
        ; The duration counts number of frames
        ; Coded values allow jumping
        ; 128 bytes *MAX* for *ALL* the sounds (there's no more room anyway :)



        ; y = sound #

        lda SoundVolume,y
        sta AUDV0
        lda SoundControl,y
        sta AUDC0
        lda SoundFrequency,y
        sta AUDF0
        rts

SOUND_OFF               = 0
SOUND_SLIDING           = 1
SOUND_JUMP              = 2
SOUND_

; interesting ones...
; 4/20/2  kind of a landing






;SOUND_COMMAND_END       = 0
;SOUND_COMMAND_BRANCH    = 255

        ; Format:  D7+[delay] | SOUND_COMMAND_BRANCH, branch
        ;       D7=0+[freq], control+volume


;        ldx #1

;TwoChanSound

;       ldy SoundPP,x
;        beq nextChannel                         ; channel inactive

;ProcessSound2

;        lda SoundProgram,y
;        bpl CFV

                ; a command

;        cmp #SOUND_COMMAND_END
;        bne notEnd
;        sta SoundPP,x                           ; halts
;        bcs nextChannel

;notEnd  cmp #SOUND_COMMAND_BRANCH
;        bne notBranch
        
;        lda SoundProgram+1,y
;        tay
;        bcs ProcessSound2

;notBranch               ; assume delay, encoded

;        and #127
;        sta SoundDelay,x
;        iny
;        bne ProcessSound2
     
;CFV
;        sta AUDF0,x
;        iny
;        lda SoundProgram,y
;        sta AUDV0,x
;        lsr
;        lsr
;        lsr
;        lsr
;        sta AUDC0,x
;        iny

;        sty SoundPP,x

;nextChannel

;        dex
;        bpl TwoChanSound

;        rts


;Sound
;        .byte 0                 ; << halted


;        .byte D7+127,0,0,SOUND_COMMAND_BRANCH,0
;        .byte 





SoundVolume             ;         ID              CONTROL VOLUME
        .byte %00000000        ; 0      QUIET
        .byte %10000001        ; 1      SLIDING          8       1
        .byte %00110010        ; 2      JUMP

SoundFrequency
        .byte 0
        .byte 31                ; 1
        .byte 10                 ; 2

SoundControl
        .byte 0
        .byte 8                 ; 1
        .byte 3                 ; 2
 

;------------------------------------------------------------------------------

LevelCount

        .byte %111,%111,%11,%11,%11,%1,%1,0

GetDifficultyMask

        ; Returns a mask for counters, based on the level
        ; The smaller the mask, the faster things happen, so increasing difficulty
        ; see 'CountDown' for flags return
        ; buggers y

        lda level
        and #%11111
        lsr
        lsr
        tay
        lda LevelCount,y
 
                ; fall through
;-----------------^^^^^^^^^^^^-------------------------------------------------

CountDown

        ; Generic decrement of the player's counter
        ; pass x = player, a = timer mask
        ; returns carry S = counter still active (non-zero on entry)
        ; y = new value of counter
        ; C = counter halted (was 0)

        sec
        and timer
        bne NotEnd

        clc
        lda PlayerCounter,x
        and #%11111
        beq NotEnd                      ; actually, finished (0-->CC)

        dec PlayerCounter,x

        sec

NotEnd  lda PlayerCounter,x
        and #%11111
        tay
        rts

;------------------------------------------------------------------------------

CorpseNormal

        asl PlayerCube
        lsr PlayerCube                          ; clear D7 (so score stops counting down)

        lda #FRAME_SAD
        jsr SetFrame

        lda #%111111
        jsr CountDown
        bcc NextLife

        lda timer
        and #7
        bne notDemo

        lda PlayerFlags
        eor #FLAG_MIRROR
        sta PlayerFlags                         ; 'shake head'

notDemo rts


;------------------------------------------------------------------------------

CorpseDeathwatch

        jsr Interact
        bne BeginGame                           ; user intervention for restart


        lsr PlayerFlags
        lda timer
        asl
        rol PlayerFlags                         ; flash between score/lives display

                ; Just waiting until we go to demo mode

        lda #%111111
        jsr CountDown
        bcs notDemo

        jmp WarmBoot

;------------------------------------------------------------------------------

GetCubeTotal

        ; Calculate total cubes required for this level
        ; returns a = count (0-7)

        IF TESTMAXCUBE

                lda #NUMBEROFCUBES

        ELSE
                lda level
                and #%11110
                lsr
                adc #3                          ; CC!
                cmp #NUMBEROFCUBES
                bcc LimitC
                lda #NUMBEROFCUBES
LimitC
        ENDIF

        rts

;------------------------------------------------------------------------------

AILock

        asl Cubes
        sec
        ror Cubes                       ; set D7 (completed)

        lda #GREEN + %11
        sta BackColour

                ; RELY on player frame already being 0 or 1

        lda timer
        and #7
        bne nPFr

        lda PlayerFlags
        eor #2
        sta PlayerFlags

nPFr

        ; Locked, pending end of screen

        lda #%1111
        jsr CountDown

        lda PlayerCube
        bmi noBonus                     ; penalised for slowness

        lda timer+1
        beq noBonus
        dec timer+1

                ; Can't finish until bonus has been added

        ldy #$05
        lda #$00
        jsr AddSmallScore               ; 500 bonus

        ldx AI_Player
        sec

noBonus bcc NextLevel           ; (jsr/rts/jmp optimised)
        rts

;------------------------------------------------------------------------------

NextLife

        jsr LifeDown
        bcs StartNewLife                        ; player still has lives left

        lda #RED + %10
        sta BackColour

        asl Cubes
        sec
        ror Cubes                               ; use 'completed' to prevent other stuff happening

        lda #MODE_DEATHWATCH + 31
        jmp ChangeMode                          ; JSR/RTS Optimisation

;------------------------------------------------------------------------------

BeginGame

                ; Game-inits go here...

        lda #%01100000
        sta level                       ; level = 0, 3 lives

        lda #0
        sta PlayerScore
        sta PlayerScore+1
        sta PlayerScore+2

NextLevel
                ; Level-inits go here...

        inc level
        lda level
        and #%11111
        bne Unchange
        dec level
Unchange



StartNewLife

        jsr GetCubeTotal
        sta Cubes


InitCubes

          ; Called on exit from demo mode, so also inits colours

        ldy #255
        sty timer+1                     ; FORCES tableaux reset in CheckCompletion

        iny
        sty BackColour
     
        ldx #4
ClGrid  sty Grid-1,x
        dex
        bne ClGrid

                ; following relies on X=0, a=0 from above

;        sty TargetDisplay+8
;        sty TargetDisplay+9                                ; SLIDING TARGET AREA


        iny
        sty timer

;        ldx #0                         X MUST BE 0

        IF TYPE_PLAYER != 0
                lda #TYPE_PLAYER
        ELSE
        jsr ChangeCreature              ; creature ID #0
        ldx AI_Player

        inx
        lda #TYPE_LURKER
        jsr ChangeCreature

        ldx #NUMBEROFCUBES-1
        lda #D7
shiftc  sta CubeX,x                     ; dead cube
        dex
        bpl shiftc

        jsr SquirtCube                  ; inject a single cube, which the player rides
        stx IC_FedCubeIndex

        ldx #1
IPC     lda PlayerCube,x
        and #127-MASK_CUBE              ; also clears Penalty flag (D7)
        ora IC_FedCubeIndex
        sta PlayerCube,x

        jsr TieToCube

        dex
        bpl IPC

                ; Dodgy use of FC_ vars... created for this usage
                ; Prevents overlay clumsiness, but beware - this routine must be last

        lda Cubes
        and #7
        lsr
        sta FC_temp2                    ; we plonk about half of the total cubes

                ; Plonk down some random cubes

Plonk   jsr Random0to3
        ldx t5,y
        jsr Random0to3
        lda t5,y
        tay
        jsr FeedCube            ; COULD fail (over another), but we don't care

        dec FC_temp2
        bne Plonk

                ; Display lives/score section

        beq DisplayLivesScore           ; JMP/JSR/RTS optimised

t5      .byte 0,5,10,15


NextLifeTable
        .byte %00100000,%01000000,%01100000,%10000000,%10100000,%11000000,%11100000,%11100000
        .byte %00000000,%00000000,%00100000,%01000000,%01100000,%10000000,%10100000,%11000000

;------------------------------------------------------------------------------

GetLife
        lda level
        asl
        rol
        rol
        rol
        and #7
        tay
        rts

;------------------------------------------------------------------------------

LifeDown

        jsr GetLife
        ora #8
        tay
        bpl LifeIs

LifeUP  jsr GetLife

LifeIs  lda level
        and #255-MASKLIVES
        ora NextLifeTable,y
        sta level

        lda NextLifeTable,y
        cmp #1                               ; -> CS = alive, CC = dead

                ; fall through
;-----------------^^^^^^^^^^^^-------------------------------------------------

DisplayLivesScore

        lda PlayerFlags
        ora #DISPLAY_LIVES_SCORE
        sta PlayerFlags
        rts

;------------------------------------------------------------------------------

Weapon
        lda Cubes
        and #CUBES_DEMO_MODE
        bne noWeapon                    ; not in demo!

        lda INPT4
        bmi noWeapon                    ; no button pressed

        cpx #0
        bne noWeapon

        lda PlayerScore
        ora PlayerScore+1
        ora PlayerScore+2
        beq noWeapon                    ; can't fire if no 'fire points'

        inx                             ; the enemy
        lda #MODE_TOSSOFF + 10
        jsr ChangeMode                  ; switch (if possible) other creature's mode

                ; BUT... there's a score penalty (10 points)

        lda #$90
        jsr DecScore

noWeapon

        ldx AI_Player
aRTS    rts

;------------------------------------------------------------------------------

DecScore

        ; a = lo of score DECREMENT

        ldx #$99
        ldy #$99
        jmp AddScore                    ; JSR/RTS optimised

;------------------------------------------------------------------------------

AINormal
Human


        lda timer
        and #63
        cmp #7
        bcs nBlink

        lda rnd+1
        bmi nBlink

        lda #FRAME_BLINK
        bne nB2
nBlink  lda #FRAME_STAND

nB2     ldy PlayerCube
        bpl nB3                         ; not penalising score for lateness
        lda #FRAME_SAD
nB3     jsr SetFrame






        lda #%1
        jsr CountDown
        bcs aRTS

        jsr Weapon             ; Check for weapon usage

Joystick
 
;        lda #0
;        sta SWACNT              ; all input bits (IMPLICITLY SET @ STARTUP)

        lda SWCHA               ; joystick
        lsr
        lsr
        lsr
        lsr
        tay

        lda joystickdir,y
        tay

        lda Cubes
        and #CUBES_DEMO_MODE
        beq HumanMove                   ; not in demo mode?

        IF ATTRACT_MODE
                        ; attract mode just randomises background colour

                lda timer
                bne NoAttract
                lda rnd
                and #%11110111          ; not too high-intensity
                sta BackColour
NoAttract
        ENDIF

        cpy #4
        bcs NoDemo                      ; human interaction?
        jmp BeginGame                   ; JSR/RTS optimisation 

NoDemo
                ; DEMO mode in control - choose a move

        jsr Random
        and #31                         ; only 0..3 legal
        tay

HumanMove


                ; Check cube-in-motion - can't do anything if so

        jsr GetCube
        lda CubeYDir,x
        bpl UnJoy                       ; Can't do anything if cube in motion

        
        cpy #4
        bcs UnJoy                       ; no legal joystick direction


        jsr CheckMove        
        bcc UnJoy                       ; OK!

        jsr CheckJump                   ; couldn't move... might be able to jump

UnJoy

;        jmp SetMirror                   ; JSR/RTS optimised.  Handle facing Left/Right

                ; fall through
;-----------------^^^^^^^^^^^^-------------------------------------------------

SetMirror

        ; Set mirror flag for creature
        ; sets creature flag.... 0 = facing left, sprite mirrored

        ldx AI_Player
        lda PlayerFlags,x
        and #D5
        beq dontCare
        
        lda PlayerFlags,x
        and #D6
        lsr
        lsr
        sta AI_temp

        lda PlayerFlags,x
        and #255-D4
        ora AI_temp
        sta PlayerFlags,x

dontCare
Ret
        rts


;------------------------------------------------------------------------------

AITossOff       
                ; code to leap off screen here...

        lda #1
        jsr CountDown
        bcc Lurker                      ; long enough - end

        lda PlayerY,x
        adc #4
        sta PlayerY,x
        bpl Ret

                ; fall through
;-----------------^^^^^^^^^^^^-------------------------------------------------

Lurker  lda #TYPE_LURKER                ; change to lurker

                ; fall through
;-----------------^^^^^^^^^^^^-------------------------------------------------

ChangeCreature

        ; Change to another creature
        ; THIS WILL CALL THE INIT FUNCTION FOR THE CREATURE!
        ; pass  a = new type
        ;       x = creature index

        sta AI_temp

        lda PlayerCube,x
        and #255-MASK_TYPE
        ora AI_temp
        sta PlayerCube,x

        lda #MODE_INIT
        sta PlayerCounter,x             ; #0 is creature init vector

        lda PlayerFlags,x
        and #%11110001-FLAG_MIRROR      ; turn off mirror, frame=0
        sta PlayerFlags,x

                ; fall through
;-----------------^^^^^^^^^^^^-------------------------------------------------

DoAI

        ; x = creature index

        stx AI_Player

        jsr GetAIVector
        lda PlayerCounter,x
        jsr GetModeIndex

        lda (AI_Vector),y
        pha
        dey
        lda (AI_Vector),y
        pha
        rts


;------------------------------------------------------------------------------

GetCube

        ; Get cube # from player
        ; pass x = player
        ; returns A,x = cube #, y unchanged

        lda PlayerCube,x
        and #MASK_CUBE
        tax
        rts

;------------------------------------------------------------------------------

IsPlayerCube

        ; Check current creature's cube against player's cube
        ; pass x = creature
        ; returns BEQ yes, BNE no

        lda PlayerCube,x
        eor PlayerCube
        and #MASK_CUBE

;LandHo
        rts

;------------------------------------------------------------------------------

;AILand  lda #1
;        jsr CountDown
;        bcs LandHo
        
                ; fall through
;-----------------^^^^^^^^^^^^-------------------------------------------------

ShifterInit

        ; also used for Hunter,2Shoes


        lda #FRAME_STAND
        jsr SetFrame

        lda #MODE_NORMAL + 2
        jmp ChangeMode                  ; JSR/RTS optimisation


;------------------------------------------------------------------------------




TwoShoesNormal
ShifterNormal


        ; Standard creature stuff
        ; returns CS if creature NOT to do anything more

        jsr HitPlayer
        jsr GoAway
        bcs noShift                     ; changed!

        jsr GetDifficultyMask
        bcs noShift                     ; wait a bit more before doing anything

        lda #MODE_NORMAL + 3
        sta PlayerCounter,x

        lda PlayerCube,x
        and #MASK_TYPE
        sta AI_temp

                ; if the current cube is misplaced, choose a new cube to wander to

        jsr GetCube
        lda CubeYDir,x
        bpl noShift                     ; cube is moving - can't do anything

        jsr IsCorrectPlace

        ldx AI_Player

        lda AI_temp
        eor #TYPE_2SHOES
        beq cSlide

        bcs SlideAway                   ; slide away correctly placed cube
        sec
cSlide  bcc SlideAway

                ; try to jump elsewhere

JumpAway
        ldx AI_Player
        jsr GetCube
        jsr Random0to3
        jsr CheckJump
       
        bcs aShift
        ldx AI_Player
        jmp SetMirror                   ; JSR/RTS optimised
aShift

                ; Failed.  Randomly shift Anyway

        jsr Random
        cmp #248
        bcc noShift

SlideAway
        ldx AI_Player
        jsr IsPlayerCube
        beq NotNasty                    ; don't move player cube

        jsr GetCube
        jsr Random0to3
        jsr CheckMove
        bcs NotNasty

        jsr SetMirror

        lda #MODE_NORMAL + 20
        sta PlayerCounter,x                     ; still active


NotNasty
noShift
       rts


;------------------------------------------------------------------------------

GoAway

;        ldx AI_Player

                ; There's a chance the creature will just go away

        jsr Random
        cmp #254
        bcc mLikely3
ToPoof  lda #TYPE_POOF
        jsr ChangeCreature
        sec

mLikely3

        rts

;------------------------------------------------------------------------------
; TwoShoes
; Places cubes in their CORRECT positions

;TwoShoesNormal

;        jsr HitPlayer
;        jsr GoAway
;        bcs noShift                     ; changed!

;        jsr GetDifficultyMask
;        bcs noShift                     ; wait a bit more before doing anything
;        lda #MODE_NORMAL + 3
;        sta PlayerCounter,x


        ; Creature assists player by moving UNTIL in correct position, then jumping

                ; if the current cube is misplaced, choose a new cube to wander to

;        jsr GetCube
;        lda CubeYDir,x
;        bpl noShift                     ; cube is moving - can't do anything

;        jsr IsCorrectPlace
;        bcs JumpAway

;------------------------------------------------------------------------------

HitPlayer

        jsr Collide
        bcs noPhit

ShootPlayer

        ldx #0
        lda #MODE_PLAYER_KILL
        jsr ChangeMode

        ldx AI_Player
noPhit  rts

;------------------------------------------------------------------------------
; Hunter
; Bee-lines for the player

HunterNormal

        jsr HitPlayer
        jsr GoAway
        bcs Hunting

        jsr GetDifficultyMask
        bcc Hunting                     ; wait a bit more before doing anything
        lda #MODE_NORMAL+2
        sta PlayerCounter,x

        jsr GetCube
        lda CubeYDir,x
        bpl Hunting                     ; Can't do anything if cube in motion

        jsr Random
        bmi HuntY                       ; 1/2 the time, try y first

        ldx AI_Player
        lda PlayerX,x
        cmp PlayerX
        beq HuntY


        ldy #1                          ; move right
        bcc Mleft

        ldy #3                          ; move left
Mleft

AMove   ldx AI_Player
        jsr GetCube
        jsr CheckMove                   ; try to slide
        bcc Hunting2
        jmp CheckJump                   ; JSR/RTS optimised   .... try to jump

HuntY
        ldx AI_Player
        jsr GetCube
        lda CubeYDir,x
        and #15
        sta AI_temp
        ldx #0
        jsr GetCube
        lda CubeYDir,x
        and #15
        cmp AI_temp
        beq Hunting

        ldy #2
        bcs AMove
        ldy #0
        beq AMove

Hunting2


Hunting
        rts



;------------------------------------------------------------------------------
; Lurker

        ; Sits around waiting for a good cube to occupy, then changes
        ; into a creature


LurkNormal

                ; choose a new cube

        jsr Random
        and #7
        cmp #NUMBEROFCUBES
        bcs StillLurk

        tay

        eor PlayerCube
        and #MASK_CUBE
        beq StillLurk                   ; don't take player's cube!

        lda CubeX,y
        bmi StillLurk                   ; cube not present

 

        sty AIL_LurkerCube

        lda PlayerCube,x
        and #255-MASK_CUBE
        ora AIL_LurkerCube
        sta PlayerCube,x                ; occupied cube

        jsr TieToCube

                ; Begin to hatch

        lda #MODE_HATCH+5
        sta PlayerCounter,x

        lda #0
        jsr SetFrame
;        lda PlayerFlags,x
;        and #255-MASK_FRAME
;        sta PlayerFlags,x

noType
StillLurk
Hatching
noPOOF
        rts


;------------------------------------------------------------------------------

LurkHatch
                ; Display an egg hatching into the creature
                ; frames 1,2,3... hatch!
                ; Hatching objects can be chomped

        jsr Collide
        bcs noCrack                     ; not stomped on

        ldy #0
        lda #$50
        jsr AddSmallScore

        ldx AI_Player
        jmp Lurker

noCrack

        lda #0
        jsr CountDown
        bcs Hatching

        lda #MODE_HATCH+5
        sta PlayerCounter,x

        lda PlayerFlags,x
        adc #FRAMEPLUSONE
        and #MASK_FRAME

        cmp #6*2                        ; frames 1,2,3 OK... 6 signals end

        jsr SetFrame                    ; assumed not to change Carry

        bcc Hatching

                ; Here we choose the creature based upon the difficulty

        jsr Random0to3
        lda rnd+1
        and #15
        beq genC                                ; every now and then, just DO it

        lda level
        and #%11111
        asl
        asl
        asl

        sbc RTypeDifficulty,y
        bcc aFruit
        cmp rnd
        bcs genC

aFruit  ldy #4                                  ; cop-out to a fruit
genC

                ; Change to a new creature (we're hatched)

        lda RType,y


        IF FORCEHUNTER
                lda #TYPE_HUNTER
        ENDIF

        IF FORCESHIFTER
                lda #TYPE_SHIFTER
        ENDIF

        IF FORCEHELPER
                lda #TYPE_2SHOES
        ENDIF

        jmp ChangeCreature


RType

        .byte TYPE_SHIFTER
        .byte TYPE_SHIFTER
        .byte TYPE_2SHOES
        .byte TYPE_HUNTER
        .byte TYPE_FRUIT

RTypeDifficulty

        .byte 17,17,34,50


;------------------------------------------------------------------------------

POOFNormal

        jsr GetDifficultyMask
        bcc noPOOF                      ; wait a bit more before doing anything

        lda #TYPE_LURKER
        jmp ChangeCreature              ; JSR/RTS optimise

;------------------------------------------------------------------------------
; Fruit...

FruitInit

        lda #MODE_NORMAL + 31
        sta PlayerCounter,x
    
                ; fall through...
;-----------------^^^^^^^^^^^^-------------------------------------------------

SetFruitFrame

                ; Set fruit frame based on global fruit counter

        lda Cubes
        and #MASK_FRUIT
        lsr
        lsr

                ; Fall through...
;-----------------^^^^^^^^^^^^-------------------------------------------------

SetFrame

        ; Set frame of creature
        ; pass x = creature, a = frame (0-7)

        sta AI_temp
        lda PlayerFlags,x
        and #255-MASK_FRAME
        ora AI_temp
        sta PlayerFlags,x
        rts

;------------------------------------------------------------------------------

FruitNormal

;        jsr Random
;        cmp #240
;        bcc NCD

        jsr GetDifficultyMask
        bcs NCD

                ; Time has expired, so fruit 'POOFS' off

        lda Cubes
        and #255-MASK_FRUIT
        sta Cubes                               ; reset fruit counter
        jmp ToPoof


NCD

                ; Check Collision

        jsr Collide
        bcs NoHit

                ; switch to a points creature

        lda #TYPE_POINTS
        jmp ChangeCreature                      ; JSR/RTS optimised

;------------------------------------------------------------------------------

PointsInit

        jsr SetFruitFrame

                ; and now.... (leaving the frame OK)...
                ; change to display the points for the creature

        lda #MODE_NORMAL + 15
        sta PlayerCounter,x


                ; use frame for indication of points value, new display

        lda PlayerFlags,x
        and #%1110                      ; frame
        tay
        iny
        lda #0
        jsr AddSmallScore                    ; score points based on progression

        ldx AI_Player

                ; determine if an extra life is needed

        lda PlayerFlags,x
        and #%1110
        cmp #%0110                      ; extra life?
        bne NoLife

        jsr LifeUP

NoLife

                ; increment fruit progression counter
                ; this gives us our NEXT fruit type

        lda Cubes
        and #255-MASK_FRUIT
        sta AI_temp

        lda Cubes
        clc
        adc #FRUITPLUSONE
        and #MASK_FRUIT
        ora AI_temp
        sta Cubes


PointAlive
NoHit   rts


PointsNormal

                ; Hang about a bit, then die

        lda timer
        lsr
        bcc halfspeed
        inc PlayerY,x
halfspeed

        lda #%1
        jsr CountDown
        bcs PointAlive

        lda #TYPE_LURKER
        jmp ChangeCreature              ; JSR/RTS optimise


;------------------------------------------------------------------------------


Collide

        ; checks collision with player
        ; assumes creature 1 vs creature 0
        ; returns carry S = no collision, C = collision


        lda PlayerCube
        eor PlayerCube+1
        and #7
        bne NotSameC                    ; not on same cube

        lda PlayerCounter
        and #%11111

NotSameC
        cmp #1                          ; 0 = carry C -> collision, else no collision
Incomplete
        rts


;------------------------------------------------------------------------------

WrapUp

                ; If the grid is complete (checked during MoveCube) then the player is informed

        ldx MC_Complete
        bne Incomplete
        lda #MODE_PLAYER_LOCK+10

                ; fall through
;-----------------^^^^^^^^^^^^-------------------------------------------------

ChangeMode

        ; Attempt to change the mode of a creautre
        ; The creature must have a handler for the mode before a change is allowed
        ; Will not change if creature is already in the requested mode

        ; pass  x = creature
        ;       a = new mode + counter

        sta AI_temp
        eor PlayerCounter,x
        and #MASK_MODE
        beq Cannot                      ; ALREADY in this mode

        jsr GetAIVector
        lda AI_temp
        jsr GetModeIndex
        lda (AI_Vector),y
        beq Cannot                      ; null entry for that mode

        lda AI_temp
        sta PlayerCounter,x             ; actually switch mode!

Cannot  rts

;------------------------------------------------------------------------------

MoveCube


        ; Combined with previous routine CheckCompletion
        ; now checks if grid is completed, as well as drawing/moving cubes
        
        ; Erase all in a single swipe


        lda timer
        cmp #1
        bne notCritical
        lda timer+1
        bne Critical                 ; skip erase if tableaux change coming

notCritical

        ldy #PFC-1
        lda #0
        ldx #128
ZapPF   sta playfield,y
        sta playfield+PFC,y
        stx playfield+PFC+PFC,y
        dey
        bpl ZapPF

Critical

        lda Cubes
        and #7                          ; inits non-zero if cubes left to feed (ie: not complete)
        sta MC_Complete                 ; assume complete

                ; Move cubes and draw

        ldx #NUMBEROFCUBES
ShiftC2 dex
        bmi WrapUp

        lda CubeX,x
        bmi ShiftC2                     ; non-cube
        lda CubeYDir,x
        bmi stopped

        lsr
        lsr
        lsr
        lsr
        and #3
        tay                             ; 0-3 direction

        clc
        lda CubeX,x
        adc DirMoveX,y
        and #31                         ; make sure D7 doesn't get set!
        sta CubeX,x

        lda CubeYDir,x
        and #$f0
        sta MC_DirOnly

        clc
        lda CubeYDir,x
        adc DirMoveY,y
        and #$0F
        ora MC_DirOnly
        sta CubeYDir,x

        asl MC_DirOnly                  ; direction bits now compatible with PlayerFlags

        lda DirMoveX,y
        asl
        asl
        sta MC_temp

        ldy #1
Shifter 
                ; Shift ALL players on this cube

        txa
        eor PlayerCube,y
        and #7
        bne not0                                ; this cube?

        clc
        lda MC_temp
        adc PlayerX,y
        sta PlayerX,y

                ; and set the facing

        lda PlayerFlags,y
        and #255-MASK_DIRECTION
        ora MC_DirOnly
        sta PlayerFlags,y

not0    dey
        bpl Shifter

        ldy CubeX,x
        cpy #16
        bcs stopped
        lda Mod5,y
        bne stopped

        lda CubeYDir,x
        and #15
        tay
        lda Mod5,y
        bne stopped


                ; hit a gridlock boundary, so halt

        asl CubeYDir,x
        sec
        ror CubeYDir,x                  ; halt movement

        ldy #SOUND_OFF
        jsr Sound

stopped

DrawCube

        stx DC_saveX

                ; If the cube's in an incorrect spot, we haven't completed

        lda CubeYDir,x
        bpl NotComp                     ; moving cube?
        jsr IsCorrectPlace
        bcs CouldBe                     ; YES!

NotComp inc MC_Complete                 ; NOT complete

CouldBe 
        ldx DC_saveX
        jmp ShiftC2


;------------------------------------------------------------------------------



DirMoveY        .byte $fF       ;...continues...
DirMoveX        .byte 0,$01,0,$FF
DirMove5Y       .byte -CUBEXY
DirMove5X       .byte 0,CUBEXY,0,-CUBEXY


;------------------------------------------------------------------------------

AddSmallScore

        ldx #0

AddScore

        ; CANNOT USE ANY OVERLAY VARIABLES --- 

        ; add (a,y,x) to player score
        ; a = low, y = high, x = hi hi
        ; can do complimentary add (ie: subtraction)

        sed
        clc
        adc PlayerScore
        sta PlayerScore
        tya
        adc PlayerScore+1
        sta PlayerScore+1
        txa
        adc PlayerScore+2
        sta PlayerScore+2

                ; Impose upper (999999) and lower (000000) bounds...

        cmp #$99
        bcc NZ
        lda #$99
        cpx #$50
        bcc WH

        lda #0
WH      sta PlayerScore
        sta PlayerScore+1
        sta PlayerScore+2

NZ      cld

                ; Ensure score is displayed

        lda PlayerFlags
        and #255-DISPLAY_LIVES_SCORE
        sta PlayerFlags
        rts

AdjustY .byte -12,0,12,0
        IF DIAGONALS
                .byte 0,0,0,0        ; <<< FIX DIAGONALS
        ENDIF
;------------------------------------------------------------------------------
Interact

        ; returns BEQ if user interaction

        lda SWCHB
        and #3
        cmp #3

        rts


;------------------------------------------------------------------------------

MTargetPart2

;->24

;\\3
        bit 0
;//
                ; right pf

        lda playfield+PFC+PFC,x ; 4
        and #$7f                ; 2
        sta PF0                 ; 3 ->41        < 49 (@38)

        lda MTA_TargetL         ; 3
;       and MT_TargetMask       ; 
        sta PF1                 ; 3 ->47        < 55 (@44)
        lda MTA_TargetR         ; 3

;\\5
        nop
        bit 0
;//

;       and MT_TargetMask       ; 
        sta PF2                 ; 3 ->58        < 66 (@55)


        ;--------------------------
        ; SPRITE DRAW

        tya                     ; 2
        sbc SpriteEnd           ; 3
        adc #SPRITE_HEIGHT      ; 2
        bcs Draw3               ; 2(3)

        nop                     ; 2
        nop                     ; 2
        sec                     ; 2
        bcs .skipDraw3          ; 3             or branch

Draw3
        lda (Sprite),y          ; 5
        sta PLAYER0SHAPE        ; 3

.skipDraw3

        ;-->18 cycles used
        ;--------------------------

; ->76

     ; LINE 4 -------------------------------

        lda #%10000000          ; 2
        sta PF0                 ; 3
        lda playfield,x         ; 4
        sta PF1                 ; 3 ->12        < 28 (@9)
        lda playfield+PFC,x     ; 4
        sta PF2                 ; 3 ->19        < 39 (@16)

;\\6
        nop
        nop
        nop
;//

                ; right pf

        lda playfield+PFC+PFC,x ; 4
        sta PF0                 ; 3 ->40        < 49 (@37)
        lda MTA_TargetL         ; 3
        and MT_TargetMask       ; 3
        sta PF1                 ; 3 ->49        < 55 (@46)
        lda MTA_TargetR         ; 3
        and MT_TargetMask       ; 3
        sta PF2                 ; 3 ->58        < 66 (@55)


        ;--------------------------
        ; SPRITE DRAW

        tya                     ; 2
        sbc SpriteEnd+1         ; 3
        adc #SPRITE_HEIGHT      ; 2
        bcs Draw4               ; 2(3)

        nop                     ; 2
        nop                     ; 2
        sec                     ; 2
        bcs .skipDraw4          ; 3

Draw4   lda (Sprite+2),y        ; 5
        sta PLAYER1SHAPE        ; 3

.skipDraw4

        lda (ColourPlayer),y
        sta COLUP0


        ;-->18 cycles used
        ;--------------------------

;->76

     ; LINE 5 -------------------------------

        lda #%00000000          ; 2
        sta PF0                 ; 3 ->5         <    (@2)
        lda playfield,x         ; 4
        sta PF1                 ; 3 ->12        < 28 (@9)
        lda playfield+PFC,x     ; 4
        sta PF2                 ; 3 ->19        < 39 (@16)

        iny                     ; 2


;\\11 ->32
        pha
        pla
        nop
        nop
;//


                ; right pf

        lda playfield+PFC+PFC,x ; 4
        and #$7f                ; 2
        sta PF0                 ; 3 ->41        < 49 (@38)
        lda MTA_TargetL         ; 3
        sta PF1                 ; 3 ->47        < 55 (@44)
        lda MTA_TargetR         ; 3
        sta PF2                 ; 3 ->53        < 66 (@50)

        ;--------------------------
        ; SPRITE DRAW

        tya                     ; 2
        sbc SpriteEnd           ; 3
        adc #SPRITE_HEIGHT      ; 2
        bcs Draw5               ; 2(3)

        nop                     ; 2
        nop                     ; 2
        sec                     ; 2
        bcs .skipDraw5          ; 3

Draw5   lda (Sprite),y          ; 5
        sta PLAYER0SHAPE        ; 3

.skipDraw5

        ;-->18 cycles used
        ;--------------------------


        lda #%10000000          ; 2 ->0         <    (@-5)
        sta PF0                 ; 3

;->76

     ; LINE 6 -------------------------------

        lda playfield,x         ; 4
        sta PF1                 ; 3 ->7         < 28 (@4)
        lda playfield+PFC,x     ; 4
        sta PF2                 ; 3 ->14        < 39 (@11)


        ;--------------------------
        ; SPRITE DRAW

        tya                     ; 2
        sbc SpriteEnd+1         ; 3
        adc #SPRITE_HEIGHT      ; 2

;->32
        nop

                ; right pf

        lda playfield+PFC+PFC,x ; 4
        sta PF0                 ; 3 ->28        < 49 (@25)
        lda MTA_TargetL         ; 3
        and MT_TargetMask       ; 3
        sta PF1                 ; 3 ->37        < 55 (@34)
        lda MTA_TargetR         ; 3
        and MT_TargetMask       ; 3
        sta PF2                 ; 3 ->46        < 66 (@43)

;->46



        bcs Draw6               ; 2(3)

        nop                     ; 2
        nop                     ; 2
        sec                     ; 2
        bcs .skipDraw6          ; 3             or branch

Draw6   lda (Sprite+2),y        ; 5
        sta PLAYER1SHAPE        ; 3

.skipDraw6


        ;-->18 cycles used
        ;--------------------------

        iny                     ; 2

; ->59

        dex                     ; 2
        bpl MTargetArea         ; 2(3)

MTADone

;->64

         ldx #3
         jsr MBlock
         jsr MBlock3

;        lda #0
        sta PF0




        ;**************************************************
        ; WARNING: VARIABLE USAGE
        ; DO *NOT* PLACE THIS BETWEEN OVERLAY VAR USAGE

        lda PlayerFlags
        and #DISPLAY_LIVES_SCORE
        beq theScore

;------------------------------------------------------------------------------

SetupBottomScore

        ; Display lives/level if joystick button pressed


        ldy #11                 ; #icon
        sty S0
        iny                     ; &icon
        sty S4


        jsr GetLife
        iny
        sty S5

        lda level
        and #%11111

        ldx #0
        stx S1
        stx S2
        stx S3

        cmp #10
        bcc nHOK
        ldx #2

nHOK    sec
s10     sbc #10
        inc S1
        bcs s10
        adc #11
        sta S1,x

        sta WSYNC
        sta WSYNC

;------------------------------------------------------------------------------

        lda #%00101100          ; colour
        bne NowShow

theScore
        


SetupTopScore

                ; Place appropriate data in score display line

        ldx #3
inscore
        ldy PSI-1,x
        
        lda PlayerScore-1,x
        and #15
        sta S0+2,y

        lda PlayerScore-1,x
        lsr
        lsr
        lsr
        lsr
        sta S0,y

        dex
        bne inscore

                 ; Now 'remove' leading zeros

        txa                     ; now 0
rlz     ora S0,x
        beq nlz
        inc S0,x
nlz     inx
        inx
        cpx #10
        bcc rlz

        inc S0+10



        lda #%01101000          ; colour

NowShow
        jsr ScoreLine           ; <<< OVERWRITES *ALL* OVERLAY VARS

        ;**************************************************

    ; make sure we have a 192 line display
    ; pal must have an even # or the colours screw up?!

;EOScr   lda INTIM
;        bne EOScr

;        sta WSYNC

        lda #D6+D1
        sta VBLANK              ; end of screen - enter blanking
    
        jmp overscan




MTargetArea

        ; Y = scanline (actually, 1/2 scanline... index to sprite graphics)
        ; X = playfield index (0-19)
        ; X incremented by 1, Y incremented by 3


     ; LINE 1 -------------------------------

        sta WSYNC               ; 3                             wait for start of hblank

        lda #%00000000          ; 2
        sta PF0                 ; 3 ->8         <    (@5)
        lda playfield,x         ; 4
        sta PF1                 ; 3 ->15        < 28 (@12)
        lda playfield+PFC,x     ; 4
        sta PF2                 ; 3 ->22        < 39 (@19)

        stx DS_temp             ; 3

                ; right pf

        lda playfield+PFC+PFC,x ; 4
        and #$7f                ; 2
        sta PF0                 ; 3 ->34        < 49 (@31)


        lda TargetIndex,x       ; 3
        tax                     ; 2
        lda 0,x                 ; 4
        sta MTA_TargetL         ; 3
;       and MT_TargetMask       ;
        sta PF1                 ; 3 ->49        < 55 (@46)
        lda 1,x                 ; 4
        sta MTA_TargetR         ; 3
;       and MT_TargetMask       ;
        sta PF2                 ; 3 ->59        < 66 (@56)

        ;--------------------------
        ; SPRITE DRAW


        tya                     ; 2
        sec                     ; 2
        sbc SpriteEnd           ; 3
        adc #SPRITE_HEIGHT      ; 2                                     || NOMINAL LINE CHANGE
        bcs Draw1               ; 2(3)

        nop                     ; 2
        nop                     ; 2
        sec                     ; 2
        bcs .skipDraw           ; 3

Draw1   lda (Sprite),y          ; 5
        sta PLAYER0SHAPE        ; 3

.skipDraw

        ;-->18 cycles used
        ;--------------------------

; = 77

     ; LINE 2 -------------------------------

                ; OVER BY 1

        lda #%10000000          ; 2 ->3
        sta PF0                 ; 3
 
        ldx DS_temp             ; 3

        lda playfield,x         ; 4
        sta PF1                 ; 3 ->16        < 28 (@13)
        lda playfield+PFC,x     ; 4
        sta PF2                 ; 3 ->23        < 39 (@20)

;\\3
;        bit 0
;//
                ; right pf


        lda MTargetMaskInit,x   ; 4
        sta MT_TargetMask       ; 3

        lda playfield+PFC+PFC,x ; 4
        sta PF0                 ; 3 ->40        < 49 (@37)
        lda MTA_TargetL         ; 3
        and MT_TargetMask       ; 3
        sta PF1                 ; 3 ->49        < 55 (@46)
        lda MTA_TargetR         ; 3
        and MT_TargetMask       ; 3
        sta PF2                 ; 3 ->58        < 66 (@55)

        ;--------------------------
        ; SPRITE DRAW

        tya                     ; 2
        sbc SpriteEnd+1         ; 3
        adc #SPRITE_HEIGHT      ; 2
        bcs Draw2               ; 2(3)

        nop                     ; 2
        nop                     ; 2
        sec                     ; 2
        bcs .skipDraw2          ; 3             or branch

Draw2   lda (Sprite+2),y        ; 5
        sta PLAYER1SHAPE        ; 3

.skipDraw2

        lda (ColourPlayer),y
        sta COLUP0


        ;-->18 cycles used
        ;--------------------------

;->76

     ; LINE 3 -------------------------------

        lda #0                  ; 2
        sta PF0                 ; 3 ->5              (@2)
        lda playfield,x         ; 4
        sta PF1                 ; 3 ->12        < 28 (@9)
        lda playfield+PFC,x     ; 4
        sta PF2                 ; 3 ->19        < 39 (@16)

        iny                     ; 2                             NEXT LINE FOR SPRITES

;\\15

        jmp MTargetPart2        ; 3


NextX   .byte 1,2,3,0

;------------------------------------------------------------------------------



        org $FF64


; Player colour *MUST* be positioned to avoid page boundary crossing when
; accessed through (),y addressing, with effectively unknown Y

PlayerColour

        IF COMPILE_VERSION = PAL


                IF PLAYER_WEARS_HAT
                        .byte %01100010
                        .byte %01100010
                        .byte %01100010
                ELSE
                        .byte %00011100
                        .byte %00011100
                        .byte %00011100 ; << hat  (TIMING PRECLUDES ACCURATE USE)
                ENDIF

                .byte %00011100
                .byte %00011100
                .byte %00011100
                .byte %00011100
                .byte %00011100 ; << body

                .byte %11011011 ; << pants

                .byte %01100010
                .byte %01100010 ; << boots
                .byte %00001010
        ELSE

                IF PLAYER_WEARS_HAT
                        .byte %01000010
                        .byte %01000010
                        .byte %01000010
                ELSE
                        .byte %00001100
                        .byte %00001100
                        .byte %00001100 ; << hat  (TIMING PRECLUDES ACCURATE USE)
                ENDIF

                .byte %00001100
                .byte %00001100
                .byte %00001100
                .byte %00001100
                .byte %00001100

                .byte %10011010 ; << pants

                .byte %01000010
                .byte %01000100 ; << boots
                .byte %00001100

        ENDIF

;------------------------------------------------------------------------------

ScoreLine

        ; WARNING:
        ; DIGIT DATA CANNOT CROSS PAGE BOUNDARY!  NEITHER CAN THE INNER LOOP!

        ; pass the 6 'digits' in S0-S5 inclusive
        ; mangled on return


        tsx
        stx stkp

        IF ATTRACT_MODE
                eor BackColour
        ENDIF

        sta COLUP0
        sta COLUP1




                ; Convert the digits S0-S5 to pointers to graphics

        ldx #12
initscore
        
        lda S0-2,x
        asl
        asl
        asl
        adc #<DigitNull
        sta S0-2,x
        lda #>DigitNull
;        adc #0                 ; << ITS A GIVEN THAT DIGITS DON'T CROSS PAGE
        sta S0-1,x

        dex
        dex
        bne initscore
 
        stx REFP0
        stx REFP1


        lda #3
        sta NUSIZ0
        sta NUSIZ1              ; 3 copies close

        sta WSYNC

;\\ NOT TO CROSS PAGE (35 cycles)
        ldy #6
D35     dey
        bne D35
        nop
        nop
        ; --> 1 + 5*y
;//

        sta RESP0
        sta WSYNC

;\\ NOT TO CROSS PAGE (36 cycles) +1
        ldy #7
D36     dey
        bne D36
;//

        sta.w RESP1                     ; <-- +1
           
        sta HMCLR
        lda #%10100000                  ; "2 more than below"?
        sta HMP1
        lda #%11000000                  ;\-->
        sta HMP0
        sta WSYNC
        sta HMOVE

        sta WSYNC


;\\ NOT TO CROSS PAGE (56 cycles) +1
        ldx #11
pause1  dex
        bne pause1
;//

        ; Tomas Jentzsch magic...

        ldy #14
        sty.w loop                      ; <-- +1

mainloop

        lda loop
        lsr
        tay

        lda (S0),y ;5
        sta PLAYER0SHAPE                ; 3 = 15

;\\3
        bit 0                   ; 3
;//

        lda (S1),y              ; 5
        sta PLAYER1SHAPE                ; 3

        IF COMPILE_ILLEGALOPCODES
        .byte $b3,S5            ; 5 lax (S5),y ;5        illegal opcode
        ELSE
        lda (S5),y
        ENDIF

        txs                     ; 2
        lda (S2),y              ; 5
        sta sreg                ; 3
        
        IF COMPILE_ILLEGALOPCODES
        .byte $b3,S3            ; 5 lax (S3),y ;5        illegal opcode
        ELSE
        lda (S3),y
        ENDIF

        lda (S4),y              ; 5 = 33

        ldy sreg                ; 3
        sty PLAYER0SHAPE                ; 3
        stx PLAYER1SHAPE                ; 3
        sta PLAYER0SHAPE                ; 3
        tsx                     ; 2
        stx PLAYER1SHAPE                ; 3 = 17

        dec loop
        bpl mainloop


        ldx stkp
        txs

        lda #0
        sta PLAYER0SHAPE
        sta PLAYER1SHAPE

        rts

;------------------------------------------------------------------------------



    org $FFFA
SerialNumber
    .word $00                   ; serial #
    .word start                 ; irq
    .word start                 ; reset vector

;EOF
